r"""
Canadian Forest Fire Weather Index System
=========================================

This submodule defines the :py:func:`xclim.indices.fire.fire_season`, :py:func:`xclim.indices.fire.drought_code` and
:py:func:`xclim.indices.fire.cffwis_indices` indices, which are used by the eponym indicators.
Users should read this module's documentation and the one of :py:func:`fire_weather_ufunc`. They should also consult the
information available at :cite:t:`code-natural_resources_canada_data_nodate`.

First adapted from Matlab code `CalcFWITimeSeriesWithStartup.m` from GFWED :cite:p:`fire-wang_updated_2015` made
for using MERRA2 data, which was a translation of FWI.vba of the Canadian Fire Weather Index system. Then, updated and
synchronized with the R code of the cffdrs package. When given the correct parameters, the current code has an error
below 3% when compared with the :cite:t:`fire-field_development_2015` data. The cffdrs R package is different from the
original 1982 implementation, and so is xclim.

Parts of the code and of the documentation in this submodule are directly taken from :cite:t:`code-cantin_canadian_2014`
which was published with the GPLv2 license.

Fire season
-----------
Fire weather indexes are iteratively computed, each day's value depending on the previous day indexes.
Additionally and optionally, the codes are "shut down" (set to NaN) in winter. There are a few ways of computing this
shut down and the subsequent spring start-up. The `fire_season` function allows for full control of that,
replicating the `fireSeason` method in the R package. It produces a mask to be given a `season_mask` in the
indicators. However, the `fire_weather_ufunc` and the indicators also accept a `season_method` parameter so the
fire season can be computed inside the iterator. Passing `season_method=None` switches to an "always on" mode
replicating the `fire` method of the R package.

The fire season determination is based on three consecutive daily maximum temperature thresholds
:cite:p:`fire-wotton_length_1993,fire-lawson_weather_2008`. A "GFWED" method is also implemented. There, the 12h
LST temperature is used instead of the daily maximum. The current implementation is slightly different from the
description in :cite:t:`fire-field_development_2015`, but it replicates the Matlab code when `temp_start_thresh` and
`temp_end_thresh` are both set to 6 degC. In xclim, the number of consecutive days, the start and end temperature
thresholds and the snow depth threshold can all be modified.

Overwintering
-------------
Additionally, overwintering of the drought code is also directly implemented in :py:func:`fire_weather_ufunc`.
The last drought_code of the season is kept in "winter" (where the fire season mask is False) and the precipitation
is accumulated until the start of the next season. The first drought code is computed as a function of these instead
of using the default DCStart value. Parameters to :py:func:`_overwintering_drought_code` are listed below.
The code for the overwintering is based on
:cite:t:`drought-mcelhinny_high-resolution_2020,drought-van_wagner_drought_1985`.

Finally, a mechanism for dry spring starts is implemented. For now, it is slightly different from what the GFWED, uses,
but seems to agree with the state of the science of the CFS. When activated, the drought code and Duff-moisture codes
are started in spring with a value that is function of the number of days since the last significant precipitation event.
The conventional start value increased by that number of days times a "dry start" factor. Parameters are controlled in
the call of the indices and :py:func:`fire_weather_ufunc`. Overwintering of the drought code overrides this mechanism if
both are activated. GFWED use a more complex approach with an added check on the previous day's snow cover for
determining "dry" points. Moreover, there, the start values are only the multiplication of a factor to the number of dry
days.

Examples
--------
The current literature seems to agree that climate-oriented series of the fire weather indexes should be computed
using only the longest fire season of each year and activating the overwintering of the drought code and the "dry
start" for the duff-moisture code. The following example uses reasonable parameters when computing over all of Canada.

.. note::

    Here the example snippets use the _indices_ defined in this very module, but we always recommend using the
    _indicators_ defined in the :py:mod:`xclim.atmos` module.

>>> ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
>>> ds = ds.assign(
...     hurs=xclim.atmos.relative_humidity_from_dewpoint(ds=ds),
...     tas=xclim.core.units.convert_units_to(ds.tas, "degC"),
...     pr=xclim.core.units.convert_units_to(ds.pr, "mm/d"),
...     sfcWind=xclim.atmos.wind_speed_from_vector(ds=ds)[0],
... )
>>> season_mask = fire_season(
...     tas=ds.tas,
...     method="WF93",
...     freq="YS",
...     # Parameters below are at their default values, but listed here for explicitness.
...     temp_start_thresh="12 degC",
...     temp_end_thresh="5 degC",
...     temp_condition_days=3,
... )
>>> out_fwi = cffwis_indices(
...     tas=ds.tas,
...     pr=ds.pr,
...     hurs=ds.hurs,
...     sfcWind=ds.sfcWind,
...     lat=ds.lat,
...     season_mask=season_mask,
...     overwintering=True,
...     dry_start="CFS",
...     prec_thresh="1.5 mm/d",
...     dmc_dry_factor=1.2,
...     # Parameters below are at their default values, but listed here for explicitness.
...     carry_over_fraction=0.75,
...     wetting_efficiency_fraction=0.75,
...     dc_start=15,
...     dmc_start=6,
...     ffmc_start=85,
... )

Similarly, the next lines calculate the fire weather indexes, but according to the parameters and options
used in NASA's GFWED datasets. Here, no need to split the fire season mask from the rest of the computation
as _all_ seasons are used, even the very short shoulder seasons.

>>> ds = open_dataset("FWI/GFWED_sample_2017.nc")
>>> out_fwi = cffwis_indices(
...     tas=ds.tas,
...     pr=ds.prbc,
...     snd=ds.snow_depth,
...     hurs=ds.rh,
...     sfcWind=ds.sfcwind,
...     lat=ds.lat,
...     season_method="GFWED",
...     overwintering=False,
...     dry_start="GFWED",
...     temp_start_thresh="6 degC",
...     temp_end_thresh="6 degC",
...     # Parameters below are at their default values, but listed here for explicitness.
...     temp_condition_days=3,
...     snow_condition_days=3,
...     dc_start=15,
...     dmc_start=6,
...     ffmc_start=85,
...     dmc_dry_factor=2,
... )
"""

# This file is structured in the following way:
# Section 1: individual codes, numba-accelerated and vectorized functions.
# Section 2: Larger computing functions (the FWI iterator and the fire_season iterator)
# Section 3: Exposed methods and indices.
#
# Methods starting with a "_" are not usable with xarray objects, whereas the others are.
from __future__ import annotations

from collections import OrderedDict
from collections.abc import Sequence

import numpy as np
import xarray as xr
from numba import njit, vectorize

from xclim.core._types import Quantified
from xclim.core.units import convert_units_to, declare_units
from xclim.indices import run_length as rl

__all__ = [
    "DAY_LENGTHS",
    "DAY_LENGTH_FACTORS",
    "build_up_index",
    "cffwis_indices",
    "daily_severity_rating",
    "drought_code",
    "duff_moisture_code",
    "fire_season",
    "fire_weather_index",
    "fire_weather_ufunc",
    "initial_spread_index",
    "overwintering_drought_code",
]

default_params: dict[str, int | float | tuple[float, str]] = {
    "temp_start_thresh": (12.0, "degC"),
    "temp_end_thresh": (5.0, "degC"),
    "snow_thresh": (0.01, "m"),
    "temp_condition_days": 3,
    "snow_condition_days": 3,
    "carry_over_fraction": 0.75,
    "wetting_efficiency_fraction": 0.75,
    "dc_start": 15,
    "dmc_start": 6,
    "ffmc_start": 85,
    "prec_thresh": (1.0, "mm/d"),
    "dc_dry_factor": 5,
    "dmc_dry_factor": 2,
    "snow_cover_days": 60,
    "snow_min_cover_frac": 0.75,
    "snow_min_mean_depth": (0.1, "m"),
}
"""
Default values for numerical parameters of fire_weather_ufunc.

Parameters with units are given as a tuple of default value and units.
A more complete explanation of these parameters is given in the doc of :py:func:`fire_weather_ufunc`.
"""

# SECTION 1 - Codes - Numba accelerated and vectorized functions

# Values taken from GFWED code
DAY_LENGTHS = np.array(
    [
        [11.5, 10.5, 9.2, 7.9, 6.8, 6.2, 6.5, 7.4, 8.7, 10, 11.2, 11.8],
        [10.1, 9.6, 9.1, 8.5, 8.1, 7.8, 7.9, 8.3, 8.9, 9.4, 9.9, 10.2],
        12 * [9],
        [7.9, 8.4, 8.9, 9.5, 9.9, 10.2, 10.1, 9.7, 9.1, 8.6, 8.1, 7.8],
        [6.5, 7.5, 9, 12.8, 13.9, 13.9, 12.4, 10.9, 9.4, 8, 7, 6],
    ]
)

DAY_LENGTH_FACTORS = np.array(
    [
        [6.4, 5.0, 2.4, 0.4, -1.6, -1.6, -1.6, -1.6, -1.6, 0.9, 3.8, 5.8],
        12 * [1.39],
        [-1.6, -1.6, -1.6, 0.9, 3.8, 5.8, 6.4, 5.0, 2.4, 0.4, -1.6, -1.6],
    ]
)


@njit
def _day_length(lat: int | float, mth: int):  # pragma: no cover
    """Return the average day length for a month within latitudinal bounds."""
    if -30 > lat >= -90:
        dl = DAY_LENGTHS[0, :]
    elif -15 > lat >= -30:
        dl = DAY_LENGTHS[1, :]
    elif 15 > lat >= -15:
        return 9
    elif 30 > lat >= 15:
        dl = DAY_LENGTHS[3, :]
    elif 90 >= lat >= 30:
        dl = DAY_LENGTHS[4, :]
    elif lat > 90 or lat < -90:
        raise ValueError("Invalid lat specified.")
    else:
        raise ValueError
    return dl[mth - 1]


@njit
def _day_length_factor(lat: float, mth: int):  # pragma: no cover
    """Return the day length factor."""
    if -15 > lat >= -90:
        dlf = DAY_LENGTH_FACTORS[0, :]
    elif 15 > lat >= -15:
        return 1.39
    elif 90 >= lat >= 15:
        dlf = DAY_LENGTH_FACTORS[2, :]
    elif lat > 90 or lat < -90:
        raise ValueError("Invalid lat specified.")
    else:
        raise ValueError
    return dlf[mth - 1]


@vectorize(nopython=True)
def _fine_fuel_moisture_code(t, p, w, h, ffmc0):  # pragma: no cover
    """Compute the fine fuel moisture code over one time step.

    Parameters
    ----------
    t : array_like
        Noon temperature [C].
    p : array_like
        Rain fall in open over previous 24 hours, at noon [mm].
    w : array_like
        Noon wind speed [km/h].
    h : array_like
        Noon relative humidity [%].
    ffmc0 : array_like
        Previous value of the fine fuel moisture code.

    Returns
    -------
    array_like
        Fine fuel moisture code at the current timestep.
    """
    mo = (147.2 * (101.0 - ffmc0)) / (59.5 + ffmc0)  # *Eq.1*#
    if p > 0.5:
        rf = p - 0.5  # *Eq.2*#
        if mo > 150.0:
            mo = (
                mo
                + 42.5 * rf * np.exp(-100.0 / (251.0 - mo)) * (1.0 - np.exp(-6.93 / rf))
            ) + (0.0015 * (mo - 150.0) ** 2) * np.sqrt(rf)
            # *Eq.3b*#
        elif mo <= 150.0:
            mo = mo + 42.5 * rf * np.exp(-100.0 / (251.0 - mo)) * (
                1.0 - np.exp(-6.93 / rf)
            )
            # *Eq.3a*#
        mo = min(mo, 250.0)

    ed = (
        0.942 * (h**0.679)
        + (11.0 * np.exp((h - 100.0) / 10.0))
        + 0.18 * (21.1 - t) * (1.0 - 1.0 / np.exp(0.1150 * h))
    )  # *Eq.4*#

    if mo < ed:
        ew = (
            0.618 * (h**0.753)
            + (10.0 * np.exp((h - 100.0) / 10.0))
            + 0.18 * (21.1 - t) * (1.0 - 1.0 / np.exp(0.115 * h))
        )  # *Eq.5*#
        if mo < ew:
            # *Eq.7a*#
            kl = 0.424 * (1.0 - ((100.0 - h) / 100.0) ** 1.7) + (
                0.0694 * np.sqrt(w)
            ) * (1.0 - ((100.0 - h) / 100.0) ** 8)
            kw = kl * (0.581 * np.exp(0.0365 * t))  # *Eq.7b*#
            m = ew - (ew - mo) / 10.0**kw  # *Eq.9*#
        elif mo > ew:
            m = mo
        else:
            raise ValueError()
    elif mo == ed:
        m = mo
    else:
        kl = 0.424 * (1.0 - (h / 100.0) ** 1.7) + (0.0694 * np.sqrt(w)) * (
            1.0 - (h / 100.0) ** 8
        )  # *Eq.6a*#
        kw = kl * (0.581 * np.exp(0.0365 * t))  # *Eq.6b*#
        m = ed + (mo - ed) / 10.0**kw  # *Eq.8*#

    ffmc = (59.5 * (250.0 - m)) / (147.2 + m)  # *Eq.10*#
    if ffmc > 101.0:
        ffmc = 101.0
    elif ffmc <= 0.0:
        ffmc = 0.0

    return ffmc


@vectorize(nopython=True)
def _duff_moisture_code(
    t: np.ndarray,
    p: np.ndarray,
    h: np.ndarray,
    mth: int,
    lat: float,
    dmc0: float,
):  # pragma: no cover
    """Compute the Duff moisture code over one time step.

    Parameters
    ----------
    t : array_like
        Noon temperature [C].
    p : array_like
        Rain fall in open over previous 24 hours, at noon [mm].
    h : array_like
        Noon relative humidity [%].
    mth : array_like[int]
        Month of the year [1-12].
    lat : float
        Latitude.
    dmc0 : float
        Previous value of the Duff moisture code.

    Returns
    -------
    array
        Duff moisture code at the current timestep
    """
    if np.isnan(dmc0):
        return np.nan

    dl = _day_length(lat, mth)

    if t < -1.1:
        rk = 0
    else:
        rk = 1.894 * (t + 1.1) * (100.0 - h) * dl * 0.0001  # *Eqs.16 and 17*#

    if p > 1.5:
        ra = p
        rw = 0.92 * ra - 1.27  # *Eq.11*#
        wmi = 20.0 + 280.0 / np.exp(
            0.023 * dmc0
        )  # *Eq.12*#  This line replicates cffdrs (R code from CFS)
        # wmi = 20.0 + np.exp(5.6348 - dmc0 / 43.43)  # *Eq.12*# This line replicates GFWED (Matlab code)
        if dmc0 <= 33.0:
            b = 100.0 / (0.5 + 0.3 * dmc0)  # *Eq.13a*#
        else:
            if dmc0 <= 65.0:
                b = 14.0 - 1.3 * np.log(dmc0)  # *Eq.13b*#
            else:
                b = 6.2 * np.log(dmc0) - 17.2  # *Eq.13c*#
        wmr = wmi + (1000 * rw) / (48.77 + b * rw)  # *Eq.14*#
        pr = 43.43 * (5.6348 - np.log(wmr - 20.0))  # *Eq.15*# cffdrs R cfs
        # pr = 244.72 - 43.43 * np.log(wmr - 20.0)  # *Eq.15*# GFWED Matlab
    else:  # p <= 1.5
        pr = dmc0

    pr = max(pr, 0.0)
    dmc = pr + rk
    dmc = max(dmc, 0.0)
    return dmc


@vectorize(nopython=True)
def _drought_code(  # pragma: no cover
    t: np.ndarray,
    p: np.ndarray,
    mth: np.ndarray,
    lat: float,
    dc0: float,
) -> np.ndarray:
    """Compute the drought code over one time step.

    Parameters
    ----------
    t : array-like
        Noon temperature [C].
    p : array_like
        Rain fall in open over previous 24 hours, at noon [mm].
    mth : array_like[int]
        Month of the year [1-12].
    lat : float
        Latitude.
    dc0 : float
        Previous value of the drought code.

    Returns
    -------
    array_like
        Drought code at the current timestep
    """
    fl = _day_length_factor(lat, mth)  # type: ignore

    t = max(t, -2.8)  # type: ignore
    pe = (0.36 * (t + 2.8) + fl) / 2  # *Eq.22*#
    pe = max(pe, 0.0)  # type: ignore

    if p > 2.8:
        ra = p
        rw = 0.83 * ra - 1.27  # *Eq.18*#  Rd
        smi = 800.0 * np.exp(-dc0 / 400.0)  # *Eq.19*# Qo
        dr = dc0 - 400.0 * np.log(1.0 + ((3.937 * rw) / smi))  # *Eqs. 20 and 21*#
        if dr > 0.0:
            dc = dr + pe
        elif np.isnan(dc0):
            dc = np.nan
        else:
            dc = pe
    else:  # f p <= 2.8:
        dc = dc0 + pe
    return dc  # type: ignore


def initial_spread_index(ws: np.ndarray, ffmc: np.ndarray) -> np.ndarray:
    """Initialize spread index.

    Parameters
    ----------
    ws : array_like
        Noon wind speed [km/h].
    ffmc : array_like
        Fine fuel moisture code.

    Returns
    -------
    array_like
        Initial spread index.
    """
    mo = 147.2 * (101.0 - ffmc) / (59.5 + ffmc)  # *Eq.1*#
    ff = 19.1152 * np.exp(mo * -0.1386) * (1.0 + (mo**5.31) / 49300000.0)  # *Eq.25*#
    isi: np.ndarray = ff * np.exp(0.05039 * ws)  # *Eq.26*#
    return isi


def build_up_index(dmc, dc):
    """Build-up index.

    Parameters
    ----------
    dmc : array
        Duff moisture code.
    dc : array
        Drought code.

    Returns
    -------
    array
        Build up index.
    """
    bui = np.where(
        dmc <= 0.4 * dc,
        (0.8 * dc * dmc) / (dmc + 0.4 * dc),  # *Eq.27a*#
        dmc - (1.0 - 0.8 * dc / (dmc + 0.4 * dc)) * (0.92 + (0.0114 * dmc) ** 1.7),
    )  # *Eq.27b*#
    return np.clip(bui, 0, None)


# TODO: Does this need to be renamed?
def fire_weather_index(isi, bui):
    """Fire weather index.

    Parameters
    ----------
    isi : array
        Initial spread index
    bui : array
        Build up index.

    Returns
    -------
    array
        Build up index.
    """
    fwi = np.where(
        bui <= 80.0,
        0.1 * isi * (0.626 * bui**0.809 + 2.0),  # *Eq.28a*#
        0.1 * isi * (1000.0 / (25.0 + 108.64 / np.exp(0.023 * bui))),
    )  # *Eq.28b*#
    fwi[fwi > 1] = np.exp(2.72 * (0.434 * np.log(fwi[fwi > 1])) ** 0.647)  # *Eq.30b*#
    return fwi


def daily_severity_rating(fwi: np.ndarray) -> np.ndarray:
    """Daily severity rating.

    Parameters
    ----------
    fwi : array_like
        Fire weather index

    Returns
    -------
    array_like
        Daily severity rating.
    """
    return 0.0272 * fwi**1.77


@vectorize(nopython=True)
def _overwintering_drought_code(DCf, wpr, a, b, minDC):  # pragma: no cover
    """Compute the season-starting drought code based on the previous season's last drought code and the total winter precipitation.

    Parameters
    ----------
    DCf : ndarray
        The previous season's last drought code
    wpr : ndarray
        The accumulated precipitation since the end of the fire season.
    a : float
        The carryover fraction from the previous season.
    b : float
        The wetting efficiency fraction.
    minDC : int
        The overwintered DC cannot be below this value, usually the normal "dc_start" value.
    """
    if np.isnan(DCf) or np.isnan(wpr):
        return np.nan
    Qf = 800 * np.exp(-DCf / 400)
    Qs = a * Qf + b * (3.94 * wpr)
    DCs = 400 * np.log(800 / Qs)
    DCs = max(DCs, minDC)
    return DCs


# SECTION 2 : Iterators


# FIXME: default_params should be supplied within the logic of the function.
def _fire_season(
    tas: np.ndarray,
    snd: np.ndarray | None = None,
    method: str = "WF93",
    temp_start_thresh: float = default_params["temp_start_thresh"][0],
    temp_end_thresh: float = default_params["temp_end_thresh"][0],
    temp_condition_days: int = default_params["temp_condition_days"],
    snow_condition_days: int = default_params["snow_condition_days"],
    snow_thresh: float = default_params["snow_thresh"][0],
) -> np.ndarray:
    """Compute the active fire season mask.

    Parameters
    ----------
    tas : ndarray
        Temperature [degC], the time axis on the last position.
    snd : ndarray, optional
        Snow depth [m], time axis on the last position, used with method == 'LA08'.
    method : {"WF93", "LA08", "GFWED"}
        Which method to use.
    temp_start_thresh : float
        Starting temperature threshold.
    temp_end_thresh : float
        Ending temperature threshold.
    temp_condition_days : int
        The number of days' temperature condition to consider.
    snow_condition_days : int
        The number of days' snow condition to consider.
    snow_thresh : float
        Numerical parameters of the methods.

    Returns
    -------
    ndarray [bool]
        `True` where the fire season is active, same shape as tas.
    """
    season_mask = np.full_like(tas, False, dtype=bool)

    if method == "WF93":
        # In WF93, the check is done the N last days, EXCLUDING the current one.
        start_index = temp_condition_days + 1
    elif method in ["LA08", "GFWED"]:
        # In LA08, the check INCLUDES the current day,
        start_index = max(temp_condition_days, snow_condition_days)
    else:
        raise ValueError("`method` must be one of 'WF93', 'LA08' or 'GFWED'.")

    for it in range(start_index, tas.shape[-1]):
        if method == "WF93":
            temp = tas[..., it - temp_condition_days : it]

            # Start up when the last X days were all above a threshold.
            start_up = np.all(temp > temp_start_thresh, axis=-1)
            # Shut down when the last X days were all below a threshold
            shut_down = np.all(temp < temp_end_thresh, axis=-1)

        elif method == "LA08":
            snow = snd[..., it - snow_condition_days + 1 : it + 1]
            temp = tas[..., it - temp_condition_days + 1 : it + 1]

            # Start up when the last X days including today have no snow on the ground.
            start_up = np.all(snow <= snow_thresh, axis=-1)
            # Shut down when today has snow OR the last X days (including today) were all below a threshold.
            shut_down = (snd[..., it] > snow_thresh) | np.all(
                temp < temp_end_thresh, axis=-1
            )

        elif method == "GFWED":
            msnow = np.mean(snd[..., it - snow_condition_days + 1 : it + 1], axis=-1)
            mtemp = np.mean(tas[..., it - temp_condition_days + 1 : it + 1], axis=-1)

            # Start up when the last X days including today have no snow on the ground.
            start_up = (mtemp > temp_start_thresh) & (msnow < snow_thresh)

            # Shut down when mean snow OR mean temp are over/under threshold
            shut_down = (msnow >= snow_thresh) | (mtemp < temp_end_thresh)
        else:
            raise ValueError("`method` must be one of 'WF93', 'LA08' or 'GFWED'.")

        # Mask is on if the previous days was on OR is there is a start-up,  AND if it's not a shut-down,
        # Aka is off if either the previous day was or it is a shut-down.
        season_mask[..., it] = (season_mask[..., it - 1] | start_up) & ~shut_down

    return season_mask


def _fire_weather_calc(  # noqa: C901
    tas, pr, rh, ws, snd, mth, lat, season_mask, dc0, dmc0, ffmc0, winter_pr, **params
):
    """Primary function computing all Fire Weather Indexes. DO NOT CALL DIRECTLY, use `fire_weather_ufunc` instead."""
    # Dear code reader, sorry.
    outputs = params["outputs"]
    ind_prevs = {"DC": dc0.copy(), "DMC": dmc0.copy(), "FFMC": ffmc0.copy()}

    season_method = params.get("season_method")
    if season_method is None:
        # None means "always on"
        season_mask = np.full_like(tas, True, dtype=bool)
        # Start with default value
        ind_prevs["DC"][np.isnan(dc0)] = params["dc_start"]
        ind_prevs["DMC"][np.isnan(dmc0)] = params["dmc_start"]
        ind_prevs["FFMC"][np.isnan(ffmc0)] = params["ffmc_start"]
    elif season_method != "mask":
        # "mask" means it was passed as an arg. Other values are methods so we compute.
        season_mask = _fire_season(
            tas,
            snd,
            method=season_method,
            temp_start_thresh=params["temp_start_thresh"],
            temp_end_thresh=params["temp_end_thresh"],
            snow_thresh=params["snow_thresh"],
            temp_condition_days=params["temp_condition_days"],
            snow_condition_days=params["snow_condition_days"],
        )

    # Codes are only computed if they are in "outputs"
    for ind in list(ind_prevs.keys()):
        if ind not in outputs:
            ind_prevs.pop(ind)

    # Outputs as a dict for easier access, but order is important in the return
    out = OrderedDict()
    for name in outputs:
        if name == "winter_pr":
            # If winter_pr was requested, it should have been given.
            out[name] = winter_pr.copy()
        elif name == "season_mask":
            # If the mask was requested as output, put the one given or computed.
            out[name] = season_mask
        else:
            # Start with NaNs
            out[name] = np.full_like(tas, np.nan)

    # Cast the mask as integers, use smallest dtype for memory purposes. (maybe this is not impact on performance?)
    season_mask = season_mask.astype(np.int16)

    overwintering = params["overwintering"]
    dry_start = params["dry_start"]

    ow_DC = dc0.copy() if dc0 is not None else None
    ow_DMC = dmc0.copy() if dmc0 is not None else None

    if overwintering and "DC" in ind_prevs:
        # In overwintering, dc0 is understood as the previous season's last DC code.
        ow_DC = dc0.copy()
        ind_prevs["DC"] = np.full_like(dc0, np.nan)

    if dry_start:
        ow_DC = dc0.copy()
        ow_DMC = dmc0.copy()
        start_up_wet = np.zeros_like(
            dmc0, dtype=bool
        )  # Pre allocate to avoid "unboundlocalerror"

    # Iterate on all days.
    for it in range(tas.shape[-1]):
        if season_method is not None:
            # Not in the always on mode, thus we must care about start-up and shut-downs of the fire season.
            if it == 0:
                if params["initial_start_up"]:
                    # As if the previous iteration was all 0s
                    delta = season_mask[..., it]
                else:
                    # Continue the previous state
                    # Meant for special corner cases when we use the season mask
                    # but some points are already "on" on the first day AND we know previous DC DMC and FFMC.
                    delta = 0 * season_mask[..., it]
            else:
                delta = season_mask[..., it] - season_mask[..., it - 1]

            # In [ME19], there are the 4 cases (in order), no need for the last one, it is implicit.
            shut_down = delta == -1
            winter = (delta == 0) & (season_mask[..., it] == 0)
            start_up = delta == 1
            # active_season = (delta == 0) & (season_mask[it] == 1)

            if dry_start:
                # When we use special start values for dry cells,
                # cells where the current precipitation is significant
                wetpts = pr[..., it] > params["prec_thresh"]

                if "SNOW" in dry_start and it >= params["snow_cover_days"]:
                    # This is for the GFWED mode with snow
                    snow_cover_history = snd[
                        ..., it - params["snow_cover_days"] + 1 : it + 1
                    ]
                    snow_days = np.count_nonzero(
                        snow_cover_history > params["snow_thresh"], axis=-1
                    )

                    # Points where the snow cover is enough to trigger a "wet" start-up.
                    start_up_wet = (
                        start_up
                        & (
                            snow_days / params["snow_cover_days"]
                            >= params["snow_min_cover_frac"]
                        )
                        & (
                            snow_cover_history.mean(axis=-1)
                            >= params["snow_min_mean_depth"]
                        )
                    )

            if "DC" in ind_prevs:
                if overwintering:
                    # Store end of season DC.
                    ow_DC[shut_down] = ind_prevs["DC"][shut_down]
                    # Fist day of winter, put current precip.
                    out["winter_pr"][shut_down] = pr[shut_down, it]
                    # Winter, add current precip.
                    out["winter_pr"][winter] = out["winter_pr"][winter] + pr[winter, it]

                    dc0 = ow_DC[start_up]
                    # Where ow_DC was NaN (happens at the start of the first season when no ow_DC was given in input),
                    # put the default start,
                    ind_prevs["DC"][start_up] = np.where(
                        np.isnan(dc0),
                        params["dc_start"],
                        _overwintering_drought_code(
                            dc0,
                            out["winter_pr"][start_up],
                            params["carry_over_fraction"],
                            params["wetting_efficiency_fraction"],
                            params["dc_start"],
                        ),
                    )
                    # Put NaN to be explicit.
                    ow_DC[start_up] = np.nan
                    out["winter_pr"][start_up] = np.nan
                elif dry_start:
                    # Dry start-up for DC is overridden by overwintering.
                    ow_DC[shut_down] = params["dc_start"]

                    if "GFWED" in dry_start:
                        # The GFWED includes the current day in the "wet points" check.
                        ow_DC[(start_up | winter) & wetpts] = 0
                        ow_DC[(start_up | winter) & ~wetpts] = (
                            ow_DC[(start_up | winter) & ~wetpts]
                            + params["dc_dry_factor"]
                        )
                    else:  # "CFS"
                        ow_DC[winter & wetpts] = params["dc_start"]
                        ow_DC[winter & ~wetpts] = (
                            ow_DC[winter & ~wetpts] + params["dc_dry_factor"]
                        )

                    if "SNOW" in dry_start:
                        # Points where we have start-up and where snow cover was enough
                        # We cancel dry dc accumulation and switch to conventional
                        ow_DC[start_up_wet] = params["dc_start"]
                    ind_prevs["DC"][start_up] = ow_DC[start_up]
                    ow_DC[start_up] = np.nan
                else:
                    ind_prevs["DC"][start_up] = params["dc_start"]
                ind_prevs["DC"][shut_down] = np.nan

            if "DMC" in ind_prevs:
                if dry_start:
                    ow_DMC[shut_down] = params["dmc_start"]

                    if "GFWED" in dry_start:
                        # The GFWED includes the current day in the "wet points" check.
                        ow_DMC[(start_up | winter) & wetpts] = 0
                        ow_DMC[(start_up | winter) & ~wetpts] = (
                            ow_DMC[(start_up | winter) & ~wetpts]
                            + params["dmc_dry_factor"]
                        )
                    else:  # "CFS"
                        ow_DMC[winter & wetpts] = params["dmc_start"]
                        ow_DMC[winter & ~wetpts] = (
                            ow_DMC[winter & ~wetpts] + params["dmc_dry_factor"]
                        )

                    if "SNOW" in dry_start:
                        # Points where we have start-up and where snow cover was enough
                        # We cancel dry dc accumulation and switch to conventional
                        ow_DMC[start_up_wet] = params["dmc_start"]
                    ind_prevs["DMC"][start_up] = ow_DMC[start_up]
                    ow_DMC[start_up] = np.nan
                else:
                    ind_prevs["DMC"][start_up] = params["dmc_start"]
                ind_prevs["DMC"][shut_down] = np.nan

            if "FFMC" in ind_prevs:
                ind_prevs["FFMC"][start_up] = params["ffmc_start"]
                ind_prevs["FFMC"][shut_down] = np.nan

        # Main computation
        if "DC" in outputs:
            out["DC"][..., it] = _drought_code(
                tas[..., it], pr[..., it], mth[..., it], lat, ind_prevs["DC"]
            )
        if "DMC" in outputs:
            out["DMC"][..., it] = _duff_moisture_code(
                tas[..., it],
                pr[..., it],
                rh[..., it],
                mth[..., it],
                lat,
                ind_prevs["DMC"],
            )
        if "FFMC" in outputs:
            with np.errstate(divide="ignore", invalid="ignore"):
                out["FFMC"][..., it] = _fine_fuel_moisture_code(
                    tas[..., it],
                    pr[..., it],
                    ws[..., it],
                    rh[..., it],
                    ind_prevs["FFMC"],
                )
        if "ISI" in outputs:
            out["ISI"][..., it] = initial_spread_index(
                ws[..., it], out["FFMC"][..., it]
            )
        if "BUI" in outputs:
            out["BUI"][..., it] = build_up_index(
                out["DMC"][..., it], out["DC"][..., it]
            )
        if "FWI" in outputs:
            out["FWI"][..., it] = fire_weather_index(
                out["ISI"][..., it], out["BUI"][..., it]
            )

        if "DSR" in outputs:
            out["DSR"][..., it] = daily_severity_rating(out["FWI"][..., it])

        # Set the previous values
        for ind, ind_prev in ind_prevs.items():
            ind_prev[...] = out[ind][..., it]

    if len(outputs) == 1:
        return out[outputs[0]]

    return tuple(out.values())


# SECTION 3 - Public methods and indices


# TODO: Does this need to be renamed?
def fire_weather_ufunc(  # noqa: C901
    *,
    tas: xr.DataArray,
    pr: xr.DataArray,
    hurs: xr.DataArray | None = None,
    sfcWind: xr.DataArray | None = None,
    snd: xr.DataArray | None = None,
    lat: xr.DataArray | None = None,
    dc0: xr.DataArray | None = None,
    dmc0: xr.DataArray | None = None,
    ffmc0: xr.DataArray | None = None,
    winter_pr: xr.DataArray | None = None,
    season_mask: xr.DataArray | None = None,
    start_dates: str | xr.DataArray | None = None,  # noqa: F841
    indexes: Sequence[str] | None = None,
    season_method: str | None = None,
    overwintering: bool = False,
    dry_start: str | None = None,
    initial_start_up: bool = True,
    **params,
) -> dict[str, xr.DataArray]:
    """Fire Weather Indexes computation using xarray's apply_ufunc.

    No unit handling. Meant to be used by power users only. Please prefer using the :py:indicator:`DC` and
    :py:indicator:`CFFWIS` indicators or the :py:func:`drought_code` and :py:func:`cffwis_indices` indices defined
    in the same submodule.

    Dask arrays must have only one chunk along the "time" dimension.
    User can control which indexes are computed with the `indexes` argument.

    Parameters
    ----------
    tas : xr.DataArray
        Noon surface temperature in °C
    pr : xr.DataArray
        Rainfall over previous 24h, at noon in mm/day
    hurs : xr.DataArray, optional
        Noon surface relative humidity in %, not needed for DC
    sfcWind : xr.DataArray, optional
        Noon surface wind speed in km/h, not needed for DC, DMC or BUI
    snd : xr.DataArray, optional
        Noon snow depth in m, only needed if `season_method` is "LA08"
    lat : xr.DataArray, optional
        Latitude in °N, not needed for FFMC or ISI
    dc0 : xr.DataArray, optional
        Previous DC map, see Notes. Defaults to NaN.
    dmc0 : xr.DataArray, optional
        Previous DMC map, see Notes. Defaults to NaN.
    ffmc0 : xr.DataArray, optional
        Previous FFMC map, see Notes. Defaults to NaN.
    winter_pr : xr.DataArray, optional
        Accumulated precipitation since the end of the last season, until the beginning of the current data, mm/day.
        Only used if `overwintering` is True, defaults to 0.
    season_mask : xr.DataArray, optional
        Boolean mask, True where/when the fire season is active.
    indexes : Sequence[str], optional
        Which indexes to compute. If intermediate indexes are needed, they will be added to the list and output.
    season_method : {None, "WF93", "LA08", "GFWED"}
        How to compute the start-up and shutdown of the fire season.
        If "None", no start-ups or shutdowns are computed, similar to the R fire function.
        Ignored if `season_mask` is given.
    overwintering : bool
        Whether to activate DC overwintering or not. If True, either season_method or season_mask must be given.
    dry_start : {None, 'CFS', 'GFWED'}
        Whether to activate the DC and DMC "dry start" mechanism and which method to use. See Notes.
        If overwintering is activated, it overrides this parameter : only DMC is handled through the dry start mechanism.
    initial_start_up : bool
        If True (default), grid points where the fire season is active on the first timestep go through a start-up phase
        for that time step. Otherwise, previous codes must be given as a continuing fire season is assumed for those points.
    carry_over_fraction : float
        Carry over fraction.
    wetting_efficiency_fraction : float
        Drought code overwintering parameters, see :py:func:`overwintering_drought_code`.
    temp_start_thresh : float
        Starting temperature threshold.
    temp_end_thresh : float
        Ending temperature threshold.
    temp_condition_days : int
        The number of days' temperature condition to consider.
    snow_thresh : float
        The snow threshold.
    snow_condition_days : int
        Parameters for the fire season determination. See :py:func:`fire_season`. Temperature is in degC, snow in m.
        The `snow_thresh` parameters is also used when `dry_start` is set to "GFWED", see Notes.
    dc_start : float
        DC start.
    dmc_start : float
        DMC start.
    ffmc_start : float
        Default starting values for the three base codes.
    prec_thresh : float
        If the "dry start" is activated, this is the "wet" day precipitation threshold, see Notes. In mm/d.
    dc_dry_factor : float
        DC's start-up values for the "dry start" mechanism, see Notes.
    dmc_dry_factor : float
        DMC's start-up values for the "dry start" mechanism, see Notes.
    snow_cover_days : int
        Snow cover days.
    snow_min_cover_frac : float
        Snow minimum cover fraction.
    snow_min_mean_depth : float
        Additional parameters for GFWED's version of the "dry start" mechanism. See Notes. Snow depth is in m.

    Returns
    -------
    dict[str, xarray.DataArray]
        Dictionary containing the computed indexes as prescribed in `indexes`, including the intermediate
        ones needed, even if they were not explicitly listed in `indexes`. When overwintering is
        activated, `winter_pr` is added. If `season_method` is not None and `season_mask` was not given,
        `season_mask` is computed on-the-fly and added to the output.

    Notes
    -----
    When overwintering is activated, the argument `dc0` is understood as last season's
    last DC map and will be used to compute the overwintered DC at the beginning of the
    next season.

    If overwintering is not activated and neither is fire season computation (`season_method`
    and `season_mask` are `None`), `dc0`, `dmc0` and `ffmc0` are understood as the codes
    on the day before the first day of FWI computation. They will default to their respective start values.
    This "always on" mode replicates the R "fire" code.

    If the "dry start" mechanism is set to "CFS" (but there is no overwintering), the arguments `dc0` and `dmc0` are
    understood as the potential start-up values from last season. With :math:`DC_{start}` the conventional start-up
    value, :math:`F_{dry-dc}` the `dc_dry_factor` and  :math:`N_{dry}` the number of days since the last significant
    precipitation event, the start-up value :math:`DC_0` is computed as:

    .. math::

        DC_0 = DC_{start} +  F_{dry-dc} * N_{dry}

    The last significant precipitation event is the last day when precipitation was greater or equal to "prec_thresh".
    The same happens for the DMC, with corresponding parameters.
    If overwintering is activated, this mechanism is only used for the DMC.

    Alternatively, `dry_start` can be set to "GFWED". In this mode, the start-up values are computed as:

    .. math::

        DC_0 = F_{dry-dc} * N_{dry}

    Where the current day is also included in the determination of :math:`N_{dry}` (:math:`DC_0` can thus be 0).
    Finally, for this "GFWED" mode, if snow cover is provided, a second check is performed: the dry start procedure is
    skipped and conventional start-up values are used for cells where the snow cover of the last `snow_cover_days` was
    above `snow_thresh` for at least `snow_cover_days` * `snow_min_cover_frac` days and where the mean snow cover over
    the same period was greater of equal to `snow_min_mean_depth`.
    """
    indexes = set(indexes or ["DC", "DMC", "FFMC", "ISI", "BUI", "FWI", "DSR"])

    if "DSR" in indexes:
        indexes.update({"FWI"})
    if "FWI" in indexes:
        indexes.update({"ISI", "BUI"})
    if "BUI" in indexes:
        indexes.update({"DC", "DMC"})
    if "ISI" in indexes:
        indexes.update({"FFMC"})
    indexes = sorted(
        list(indexes),
        key=["DC", "DMC", "FFMC", "ISI", "BUI", "FWI", "DSR"].index,
    )

    # Whether each argument is needed in _fire_weather_calc
    # Same order as _fire_weather_calc, Assumes the list of indexes is complete.
    # (name, list of indexes + start_up/shut_down modes, has_time_dim)
    needed_args = (
        (tas, "tas", ["DC", "DMC", "FFMC", "WF93", "LA08"], True),
        (pr, "pr", ["DC", "DMC", "FFMC"], True),
        (hurs, "hurs", ["DMC", "FFMC"], True),
        (sfcWind, "sfcWind", ["FFMC"], True),
        (snd, "snd", ["LA08"], True),
        (tas.time.dt.month, "month", ["DC", "DMC"], True),
        (lat, "lat", ["DC", "DMC"], False),
    )
    # Arg order : tas, pr, hurs, sfcWind, snd, mth, lat, season_mask, dc0, dmc0, ffmc0, winter_pr
    #              0   1    2      3        4   5    6    7             8    9     10    11
    args: list[xr.DataArray | None] = [None] * 12
    input_core_dims: list[list[str | None]] = [[]] * 12

    # Verification of all arguments
    for i, (arg, name, usedby, has_time_dim) in enumerate(needed_args):
        if any(ind in indexes + [season_method] for ind in usedby):
            if arg is None:
                raise TypeError(
                    f"Missing input argument {name} for index combination {indexes} "
                    f"with fire season method '{season_method}'."
                )
            args[i] = arg
            input_core_dims[i] = ["time"] if has_time_dim else []

    # For the GFWED dry start mode, we include snow depth is available
    if snd is not None and dry_start == "GFWED":
        args[4] = snd
        input_core_dims[4] = ["time"]
        dry_start = "GFWED+SNOW"
    elif dry_start not in [None, "CFS", "GFWED"]:
        raise ValueError("'dry_start' must be one of None, 'CFS' or 'GFWED'.")

    # Always pass the previous codes.
    _dc0 = xr.full_like(tas.isel(time=0), np.nan) if dc0 is None else dc0
    _dmc0 = xr.full_like(tas.isel(time=0), np.nan) if dmc0 is None else dmc0
    _ffmc0 = xr.full_like(tas.isel(time=0), np.nan) if ffmc0 is None else ffmc0
    args[8:11] = [_dc0, _dmc0, _ffmc0]

    # Output config from the current indexes list
    outputs = indexes
    output_dtypes: list[np.dtype] = [tas.dtype] * len(indexes)
    output_core_dims = len(indexes) * [("time",)]

    if season_mask is not None:
        # A mask was passed, ignore passed method and tell the ufunc to use it.
        args[7] = season_mask
        input_core_dims[7] = ["time"]
        season_method = "mask"
    elif season_method is not None:
        # Season mask not given and a method chosen : we output the computed mask.
        outputs.append("season_mask")
        output_core_dims.append(("time",))
        output_dtypes.append(bool)

    if overwintering:
        # Overwintering code activated
        if season_method is None and season_mask is None:
            raise ValueError(
                "If overwintering is activated, either `season_method` or `season_mask` must be given."
            )

        # Last winter PR is 0 by default
        if winter_pr is None:
            winter_pr = xr.zeros_like(pr.isel(time=0))

        args[11] = winter_pr

        # Activating overwintering will produce an extra output, that has no "time" dimension.
        outputs.append("winter_pr")
        output_core_dims.append([])
        output_dtypes.append(pr.dtype)

    # Kwargs from default parameters. take the value when it is a tuple.
    kwargs = {
        k: v if not isinstance(v, tuple) else v[0] for k, v in default_params.items()
    }
    kwargs.update(**params)
    kwargs.update(
        season_method=season_method,
        overwintering=overwintering,
        dry_start=dry_start,
        initial_start_up=initial_start_up,
        outputs=outputs,
    )

    if tas.ndim == 1:
        dummy_dim = xr.core.utils.get_temp_dimname(tas.dims, "dummy")  # noqa
        # When arrays only have the 'time' dimension, non-temporal inputs of the wrapped ufunc
        # become scalars. We add a dummy dimension so that we don't have to deal with that.
        for i, arg in enumerate(args):
            if isinstance(arg, xr.DataArray):
                args[i] = arg.expand_dims({dummy_dim: [1]})

    das = xr.apply_ufunc(
        _fire_weather_calc,
        *args,
        kwargs=kwargs,
        input_core_dims=input_core_dims,
        output_core_dims=output_core_dims,
        dask="parallelized",
        dask_gufunc_kwargs={
            "meta": tuple(np.array((), dtype=dtype) for dtype in output_dtypes)
        },
    )

    if tas.ndim == 1:
        if len(outputs) == 1:
            das = das.squeeze(dummy_dim, drop=True)
        else:
            das = [da.squeeze(dummy_dim, drop=True) for da in das]

    if len(outputs) == 1:
        return {outputs[0]: das}

    return dict(zip(outputs, das, strict=False))


@declare_units(last_dc="[]", winter_pr="[length]")
def overwintering_drought_code(
    last_dc: xr.DataArray,
    winter_pr: xr.DataArray,
    carry_over_fraction: xr.DataArray | float = default_params["carry_over_fraction"],
    wetting_efficiency_fraction: xr.DataArray | float = default_params[
        "wetting_efficiency_fraction"
    ],
    min_dc: xr.DataArray | float = default_params["dc_start"],
) -> xr.DataArray:
    """Compute season-starting drought code based on previous season's last drought code and total winter precipitation.

    This method replicates the "wDC" method of the "cffdrs R package :cite:p:`cantin_canadian_2014`, with an added
    control on the "minimum" DC.

    Parameters
    ----------
    last_dc : xr.DataArray
        The previous season's last drought code.
    winter_pr : xr.DataArray
        The accumulated precipitation since the end of the fire season.
    carry_over_fraction : xr.DataArray or float
        Carry-over fraction of last fall’s moisture
    wetting_efficiency_fraction : xr.DataArray or float
        Effectiveness of winter precipitation in recharging moisture reserves in spring
    min_dc : xr.DataArray or float
        Minimum drought code starting value.

    Returns
    -------
    wDC : xr.DataArray
        Overwintered drought code.

    Notes
    -----
    Details taken from the "cffdrs" R package documentation :cite:p:`drought-cantin_canadian_2014`:
    Of the three fuel moisture codes (i.e.  FFMC, DMC and DC) making up the FWI System, only the DC needs to be
    considered in terms of its values carrying over from one fire season to the next. In Canada both the FFMC and the
    DMC are assumed to reach moisture saturation from overwinter precipitation at or before spring melt; this is a
    reasonable assumption and any error in these assumed starting conditions quickly disappears. If snowfall (or other
    overwinter precipitation) is not large enough however, the fuel layer tracked by the Drought Code may not fully
    reach saturation after spring snow melt; because of the long response time in this fuel layer (53 days in standard
    conditions) a large error in this spring starting condition can affect the DC for a significant portion of the fire
    season. In areas where overwinter precipitation is 200 mm or more, full moisture recharge occurs and DC
    overwintering is usually unnecessary.  More discussion of overwintering and fuel drying time lag can be found in
    :cite:t:`drought-lawson_weather_2008` and :cite:t:`drought-van_wagner_drought_1985`.

    Carry-over fraction of last fall's moisture:
        - 1.0, Daily DC calculated up to 1 November; continuous snow cover, or freeze-up, whichever comes first
        - 0.75, Daily DC calculations stopped before any of the above conditions met or the area is subject to
          occasional winter chinook conditions, leaving the ground bare and subject to moisture depletion
        - 0.5,  Forested areas subject to long periods in fall or winter that favor depletion of soil moisture

    Effectiveness of winter precipitation in recharging moisture reserves in spring:
        - 0.9, Poorly drained, boggy sites with deep organic layers
        - 0.75, Deep ground frost does not occur until late fall, if at all; moderately drained sites that allow
          infiltration of most of the melting snowpack
        - 0.5, Chinook-prone areas and areas subject to early and deep ground frost; well-drained soils favoring rapid
          percolation or topography favoring rapid runoff before melting of ground frost

    Source: :cite:cts:`drought-lawson_weather_2008` - Table 9.

    References
    ----------
    :cite:t:`drought-cantin_canadian_2014,drought-field_development_2015,drought-lawson_weather_2008,drought-van_wagner_drought_1985`
    """
    winter_pr = convert_units_to(winter_pr, "mm")

    wDC = xr.apply_ufunc(  # noqa
        _overwintering_drought_code,
        last_dc,
        winter_pr,
        carry_over_fraction,
        wetting_efficiency_fraction,
        min_dc,
        input_core_dims=[[]] * 5,
        output_core_dims=[[]],
        dask="parallelized",
        output_dtypes=[last_dc.dtype],
    )
    wDC.attrs["units"] = ""
    return wDC


def _convert_parameters(
    params: dict[str, int | float], funcname: str = "fire weather indices"
) -> dict[str, int | float]:
    for param, value in params.copy().items():
        if param not in default_params:
            raise ValueError(
                f"{param} is not a valid parameter for {funcname}. See the docstring of the function and the list in xc.indices.fire.default_params."
            )
        if isinstance(default_params[param], tuple):
            params[param] = convert_units_to(value, default_params[param][1])
    return params


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    sfcWind="[speed]",
    hurs="[]",
    lat="[]",
    snd="[length]",
    ffmc0="[]",
    dmc0="[]",
    dc0="[]",
    season_mask="[]",
)
def cffwis_indices(
    tas: xr.DataArray,
    pr: xr.DataArray,
    sfcWind: xr.DataArray,
    hurs: xr.DataArray,
    lat: xr.DataArray,
    snd: xr.DataArray | None = None,
    ffmc0: xr.DataArray | None = None,
    dmc0: xr.DataArray | None = None,
    dc0: xr.DataArray | None = None,
    season_mask: xr.DataArray | None = None,
    season_method: str | None = None,
    overwintering: bool = False,
    dry_start: str | None = None,
    initial_start_up: bool = True,
    **params,
) -> tuple[
    xr.DataArray, xr.DataArray, xr.DataArray, xr.DataArray, xr.DataArray, xr.DataArray
]:
    """Canadian Fire Weather Index System indices.

    Computes the 6 fire weather indexes as defined by the Canadian Forest Service: the Drought Code, the Duff-Moisture
    Code, the Fine Fuel Moisture Code, the Initial Spread Index, the Build Up Index and the Fire Weather Index.

    Parameters
    ----------
    tas : xr.DataArray
        Noon temperature.
    pr : xr.DataArray
        Rain fall in open over previous 24 hours, at noon.
    sfcWind : xr.DataArray
        Noon wind speed.
    hurs : xr.DataArray
        Noon relative humidity.
    lat : xr.DataArray
        Latitude coordinate
    snd : xr.DataArray
        Noon snow depth, only used if `season_method='LA08'` is passed.
    ffmc0 : xr.DataArray
        Initial values of the fine fuel moisture code.
    dmc0 : xr.DataArray
        Initial values of the Duff moisture code.
    dc0 : xr.DataArray
        Initial values of the drought code.
    season_mask : xr.DataArray, optional
        Boolean mask, True where/when the fire season is active.
    season_method : {None, "WF93", "LA08", "GFWED"}
        How to compute the start-up and shutdown of the fire season.
        If "None", no start-ups or shutdowns are computed, similar to the R fire function.
        Ignored if `season_mask` is given.
    overwintering : bool
        Whether to activate DC overwintering or not. If True, either season_method or season_mask must be given.
    dry_start : {None, 'CFS', 'GFWED'}
        Whether to activate the DC and DMC "dry start" mechanism or not, see :py:func:`fire_weather_ufunc`.
    initial_start_up : bool
        If True (default), gridpoints where the fire season is active on the first timestep go through a start_up phase
        for that time step. Otherwise, previous codes must be given as a continuing fire season is assumed for those
        points.
    params
        Any other keyword parameters as defined in :py:func:`fire_weather_ufunc` and in :py:data:`default_params`.

    Returns
    -------
    DC: xr.DataArray, [dimensionless]
    DMC: xr.DataArray, [dimensionless]
    FFMC: xr.DataArray, [dimensionless]
    ISI: xr.DataArray, [dimensionless]
    BUI: xr.DataArray, [dimensionless]
    FWI: xr.DataArray, [dimensionless]

    Notes
    -----
    See :cite:t:`code-natural_resources_canada_data_nodate`, the :py:mod:`xclim.indices.fire` module documentation,
    and the docstring of :py:func:`fire_weather_ufunc` for more information. This algorithm follows the official R code
    released by the CFS, which contains revisions from the original 1982 Fortran code.

    References
    ----------
    :cite:cts:`fire-wang_updated_2015`
    """
    tas = convert_units_to(tas, "C")
    pr = convert_units_to(pr, "mm/day")
    sfcWind = convert_units_to(sfcWind, "km/h")
    hurs = convert_units_to(hurs, "pct")
    if snd is not None:
        snd = convert_units_to(snd, "m")

    out = fire_weather_ufunc(
        tas=tas,
        pr=pr,
        hurs=hurs,
        sfcWind=sfcWind,
        lat=lat,
        dc0=dc0,
        dmc0=dmc0,
        ffmc0=ffmc0,
        snd=snd,
        indexes=["DC", "DMC", "FFMC", "ISI", "BUI", "FWI"],
        season_mask=season_mask,
        season_method=season_method,
        overwintering=overwintering,
        dry_start=dry_start,
        initial_start_up=initial_start_up,
        **_convert_parameters(params),
    )
    for outd in out.values():
        outd.attrs["units"] = ""
    return out["DC"], out["DMC"], out["FFMC"], out["ISI"], out["BUI"], out["FWI"]


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    lat="[]",
    snd="[length]",
    dc0="[]",
    season_mask="[]",
)
def drought_code(
    tas: xr.DataArray,
    pr: xr.DataArray,
    lat: xr.DataArray,
    snd: xr.DataArray | None = None,
    dc0: xr.DataArray | None = None,
    season_mask: xr.DataArray | None = None,
    season_method: str | None = None,
    overwintering: bool = False,
    dry_start: str | None = None,
    initial_start_up: bool = True,
    **params,
) -> xr.DataArray:
    r"""Drought code (FWI component).

    The drought code is part of the Canadian Forest Fire Weather Index System.
    It is a numeric rating of the average moisture content of organic layers.

    Parameters
    ----------
    tas : xr.DataArray
        Noon temperature.
    pr : xr.DataArray
        Rain fall in open over previous 24 hours, at noon.
    lat : xr.DataArray
        Latitude coordinate
    snd : xr.DataArray
        Noon snow depth.
    dc0 : xr.DataArray
        Initial values of the drought code.
    season_mask : xr.DataArray, optional
        Boolean mask, True where/when the fire season is active.
    season_method : {None, "WF93", "LA08", "GFWED"}
        How to compute the start-up and shutdown of the fire season.
        If "None", no start-ups or shutdowns are computed, similar to the R fire function.
        Ignored if `season_mask` is given.
    overwintering : bool
        Whether to activate DC overwintering or not. If True, either season_method or season_mask must be given.
    dry_start : {None, "CFS", 'GFWED'}
        Whether to activate the DC and DMC "dry start" mechanism and which method to use.
        See :py:func:`fire_weather_ufunc`.
    initial_start_up : bool
        If True (default), grid points where the fire season is active on the first timestep go through a start_up phase
        for that time step. Otherwise, previous codes must be given as a continuing fire season is assumed for those
        points.
    params
        Any other keyword parameters as defined in `xclim.indices.fire.fire_weather_ufunc` and in :py:data:`default_params`.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Drought code

    Notes
    -----
    See :cite:cts:`code-natural_resources_canada_data_nodate`, the :py:mod:`xclim.indices.fire` module documentation,
    and the docstring of :py:func:`fire_weather_ufunc` for more information. This algorithm follows the official R code
    released by the CFS, which contains revisions from the original 1982 Fortran code.

    References
    ----------
    :cite:cts:`fire-wang_updated_2015`
    """
    tas = convert_units_to(tas, "C")
    pr = convert_units_to(pr, "mm/day")
    if snd is not None:
        snd = convert_units_to(snd, "m")

    out = fire_weather_ufunc(
        tas=tas,
        pr=pr,
        lat=lat,
        dc0=dc0,
        snd=snd,
        indexes=["DC"],
        season_mask=season_mask,
        season_method=season_method,
        overwintering=overwintering,
        dry_start=dry_start,
        initial_start_up=initial_start_up,
        **_convert_parameters(params, "drought_code"),
    )
    out["DC"].attrs["units"] = ""
    return out["DC"]


@declare_units(
    tas="[temperature]",
    pr="[precipitation]",
    hurs="[]",
    lat="[]",
    snd="[length]",
    dmc0="[]",
    season_mask="[]",
)
def duff_moisture_code(
    tas: xr.DataArray,
    pr: xr.DataArray,
    hurs: xr.DataArray,
    lat: xr.DataArray,
    snd: xr.DataArray | None = None,
    dmc0: xr.DataArray | None = None,
    season_mask: xr.DataArray | None = None,
    season_method: str | None = None,
    dry_start: str | None = None,
    initial_start_up: bool = True,
    **params,
) -> xr.DataArray:
    r"""Duff moisture code (FWI component).

    The duff moisture code is part of the Canadian Forest Fire Weather Index System.
    It is a numeric rating of the average moisture content of loosely compacted organic layers of moderate depth.

    Parameters
    ----------
    tas : xr.DataArray
        Noon temperature.
    pr : xr.DataArray
        Rain fall in open over previous 24 hours, at noon.
    hurs : xr.DataArray
        Noon relative humidity.
    lat : xr.DataArray
        Latitude coordinate
    snd : xr.DataArray
        Noon snow depth.
    dmc0 : xr.DataArray
        Initial values of the duff moisture code.
    season_mask : xr.DataArray, optional
        Boolean mask, True where/when the fire season is active.
    season_method : {None, "WF93", "LA08", "GFWED"}
        How to compute the start-up and shutdown of the fire season.
        If "None", no start-ups or shutdowns are computed, similar to the R fire function.
        Ignored if `season_mask` is given.
    dry_start : {None, "CFS", 'GFWED'}
        Whether to activate the DC and DMC "dry start" mechanism and which method to use.
        See :py:func:`fire_weather_ufunc`.
    initial_start_up : bool
        If True (default), grid points where the fire season is active on the first timestep go through a start_up phase
        for that time step. Otherwise, previous codes must be given as a continuing fire season is assumed for those
        points.
    params
        Any other keyword parameters as defined in `xclim.indices.fire.fire_weather_ufunc` and in :py:data:`default_params`.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Duff moisture code

    Notes
    -----
    See :cite:cts:`code-natural_resources_canada_data_nodate`, the :py:mod:`xclim.indices.fire` module documentation,
    and the docstring of :py:func:`fire_weather_ufunc` for more information. This algorithm follows the official R code
    released by the CFS, which contains revisions from the original 1982 Fortran code.

    References
    ----------
    :cite:cts:`fire-wang_updated_2015`
    """
    tas = convert_units_to(tas, "C")
    pr = convert_units_to(pr, "mm/day")
    if snd is not None:
        snd = convert_units_to(snd, "m")

    out = fire_weather_ufunc(
        tas=tas,
        pr=pr,
        hurs=hurs,
        lat=lat,
        dmc0=dmc0,
        snd=snd,
        indexes=["DMC"],
        season_mask=season_mask,
        season_method=season_method,
        dry_start=dry_start,
        initial_start_up=initial_start_up,
        **_convert_parameters(params, "duff_moisture_code"),
    )
    out["DMC"].attrs["units"] = ""
    return out["DMC"]


@declare_units(
    tas="[temperature]",
    snd="[length]",
    temp_start_thresh="[temperature]",
    temp_end_thresh="[temperature]",
    snow_thresh="[length]",
)
def fire_season(
    tas: xr.DataArray,
    snd: xr.DataArray | None = None,
    method: str = "WF93",
    freq: str | None = None,
    temp_start_thresh: Quantified = "12 degC",
    temp_end_thresh: Quantified = "5 degC",
    temp_condition_days: int = 3,
    snow_condition_days: int = 3,
    snow_thresh: Quantified = "0.01 m",
) -> xr.DataArray:
    """Fire season mask.

    Binary mask of the active fire season, defined by conditions on consecutive daily temperatures and, optionally, snow depths.

    Parameters
    ----------
    tas : xr.DataArray
        Daily surface temperature, cffdrs recommends using maximum daily temperature.
    snd : xr.DataArray, optional
        Snow depth, used with method == 'LA08'.
    method : {"WF93", "LA08", "GFWED"}
        Which method to use. "LA08"  and "GFWED" need the snow depth.
    freq : str, optional
        If given only the longest fire season for each period defined by this frequency,
        Every "seasons" are returned if None, including the short shoulder seasons.
    temp_start_thresh : Quantified
        Minimal temperature needed to start the season. Must be scalar.
    temp_end_thresh : Quantified
        Maximal temperature needed to end the season. Must be scalar.
    temp_condition_days : int
        Number of days with temperature above or below the thresholds to trigger a start or an end of the fire season.
    snow_condition_days : int
        Parameters for the fire season determination. See :py:func:`fire_season`. Temperature is in degC, snow in m.
        The `snow_thresh` parameters is also used when `dry_start` is set to "GFWED".
    snow_thresh : Quantified
        Minimal snow depth level to end a fire season, only used with method "LA08".
        Must be scalar.

    Returns
    -------
    xr.DataArray
        Fire season mask

    References
    ----------
    :cite:cts:`fire-wotton_length_1993,fire-lawson_weather_2008`
    """
    # TODO: `map_blocks` as currently implemented, does not mix well with non-scalar thresholds, as they are being passed through kwargs
    if not all(
        np.isscalar(v) for v in [temp_start_thresh, temp_end_thresh, snow_thresh]
    ):
        raise ValueError("Thresholds must be scalar.")

    kwargs = {
        "method": method,
        "temp_start_thresh": convert_units_to(temp_start_thresh, "degC"),
        "temp_end_thresh": convert_units_to(temp_end_thresh, "degC"),
        "temp_condition_days": temp_condition_days,
        "snow_condition_days": snow_condition_days,
        "snow_thresh": convert_units_to(snow_thresh, "m"),
    }

    def _apply_fire_season(ds, **kwargs):
        season_mask = ds.tas.copy(
            data=_fire_season(
                tas=ds.tas.values,
                snd=None if kwargs["method"] == "WF93" else ds.snd.values,
                **kwargs,
            )
        )
        season_mask.attrs = {}

        if freq is not None:
            time = season_mask.time
            season_mask = season_mask.resample(time=freq).map(rl.keep_longest_run)
            season_mask["time"] = time

        return season_mask

    ds = convert_units_to(tas, "degC").rename("tas").to_dataset()
    if snd is not None:
        ds["snd"] = convert_units_to(snd, "m")
        ds = ds.unify_chunks()
    ds = ds.transpose(..., "time")

    tmpl = xr.full_like(tas, np.nan)
    out = ds.map_blocks(_apply_fire_season, template=tmpl, kwargs=kwargs)
    out.attrs["units"] = ""
    return out



# pylint: disable=missing-kwoa
from __future__ import annotations

import jsonpickle
import numpy as np
import pytest
import xarray as xr

from xclim import set_options
from xclim.sdba.base import Grouper, Parametrizable, map_blocks, map_groups


class ATestSubClass(Parametrizable):
    pass


def test_param_class():
    gr = Grouper(group="time.month")
    in_params = dict(
        anint=4, abool=True, astring="a string", adict={"key": "val"}, group=gr
    )
    obj = Parametrizable(**in_params)

    assert obj.parameters == in_params

    assert repr(obj).startswith(
        "Parametrizable(anint=4, abool=True, astring='a string', adict={'key': 'val'}, "
        "group=Grouper("
    )

    s = jsonpickle.encode(obj)
    obj2 = jsonpickle.decode(s)  # noqa: S301
    assert obj.parameters == obj2.parameters


@pytest.mark.parametrize(
    "group,window,nvals",
    [("time", 1, 366), ("time.month", 1, 31), ("time.dayofyear", 5, 1)],
)
def test_grouper_group(tas_series, group, window, nvals):
    tas = tas_series(np.ones(366), start="2000-01-01")

    grouper = Grouper(group, window=window)
    grpd = grouper.group(tas)

    if window > 1:
        assert "window" in grpd.dims

    assert grpd.count().max() == nvals


@pytest.mark.parametrize(
    "group,interp,val90",
    [("time", False, True), ("time.month", False, 3), ("time.month", True, 3.5)],
)
def test_grouper_get_index(tas_series, group, interp, val90):
    tas = tas_series(np.ones(366), start="2000-01-01")
    grouper = Grouper(group)
    indx = grouper.get_index(tas, interp=interp)
    # 90 is March 31st
    assert indx[90] == val90


# xarray does not yet access "week" or "weekofyear" with groupby in a pandas-compatible way for cftime objects.
# See: https://github.com/pydata/xarray/discussions/6375
@pytest.mark.slow
@pytest.mark.parametrize(
    "group,n",
    [("time", 1), ("time.month", 12), ("time.week", 52)],
)
@pytest.mark.parametrize("use_dask", [True, False])
def test_grouper_apply(tas_series, use_dask, group, n):
    tas1 = tas_series(np.arange(366), start="2000-01-01")
    tas0 = tas_series(np.zeros(366), start="2000-01-01")
    tas = xr.concat((tas1, tas0), dim="lat")

    grouper = Grouper(group)
    if not group.startswith("time"):
        tas = tas.rename(time=grouper.dim)
        tas1 = tas1.rename(time=grouper.dim)
        tas0 = tas0.rename(time=grouper.dim)

    if use_dask:
        tas = tas.chunk({"lat": 1, grouper.dim: -1})
        tas0 = tas1.chunk({grouper.dim: -1})
        tas1 = tas0.chunk({grouper.dim: -1})

    # Normal monthly mean
    out_mean = grouper.apply("mean", tas)
    if grouper.prop != "group":
        exp = tas.groupby(group).mean()
    else:
        exp = tas.mean(dim=grouper.dim).expand_dims("group").T
    np.testing.assert_array_equal(out_mean, exp)

    # With additional dimension included
    grouper = Grouper(group, add_dims=["lat"])
    out = grouper.apply("mean", tas)
    assert out.ndim == 1
    np.testing.assert_array_equal(out, exp.mean("lat"))
    assert out.attrs["group"] == group
    assert out.attrs["group_compute_dims"] == [grouper.dim, "lat"]
    assert out.attrs["group_window"] == 1

    # Additional but main_only
    out = grouper.apply("mean", tas, main_only=True)
    np.testing.assert_array_equal(out, out_mean)

    # With window
    win_grouper = Grouper(group, window=5)
    out = win_grouper.apply("mean", tas)
    rolld = tas.rolling({win_grouper.dim: 5}, center=True).construct(
        window_dim="window"
    )
    if grouper.prop != "group":
        exp = rolld.groupby(group).mean(dim=[win_grouper.dim, "window"])
    else:
        exp = rolld.mean(dim=[grouper.dim, "window"]).expand_dims("group").T
    np.testing.assert_array_equal(out, exp)

    # With function + nongrouping-grouped
    grouper = Grouper(group)

    def normalize(grp, dim):
        return grp / grp.mean(dim=dim)

    normed = grouper.apply(normalize, tas)
    assert normed.shape == tas.shape
    if use_dask:
        assert normed.chunks == ((1, 1), (366,))

    # With window + nongrouping-grouped
    out = win_grouper.apply(normalize, tas)
    assert out.shape == tas.shape

    # Mixed output
    def mixed_reduce(grdds, dim=None):
        tas1 = grdds.tas1.mean(dim=dim)
        tas0 = grdds.tas0 / grdds.tas0.mean(dim=dim)
        tas1.attrs["_group_apply_reshape"] = True
        return xr.Dataset(data_vars={"tas1_mean": tas1, "norm_tas0": tas0})

    out = grouper.apply(mixed_reduce, {"tas1": tas1, "tas0": tas0})
    assert grouper.prop not in out.norm_tas0.dims
    assert grouper.prop in out.tas1_mean.dims

    if use_dask:
        assert out.tas1_mean.chunks == ((n,),)
        assert out.norm_tas0.chunks == ((366,),)

    # Mixed input
    def normalize_from_precomputed(grpds, dim=None):
        return (grpds.tas / grpds.tas1_mean).mean(dim=dim)

    out = grouper.apply(
        normalize_from_precomputed, {"tas": tas, "tas1_mean": out.tas1_mean}
    ).isel(lat=0)
    if grouper.prop == "group":
        exp = normed.mean("time").isel(lat=0)
    else:
        exp = normed.groupby(group).mean().isel(lat=0)
    assert grouper.prop in out.dims
    np.testing.assert_allclose(out, exp, rtol=1e-10)


class TestMapBlocks:
    def test_lat_lon(self, tas_series):
        tas = tas_series(np.arange(366), start="2000-01-01")
        tas = tas.expand_dims(lat=[1, 2, 3, 4]).chunk()

        # Test dim parsing
        @map_blocks(reduces=["lat"], data=["lon"])
        def func(ds, *, group, lon=None):
            assert group.window == 5
            d = ds.tas.rename(lat="lon")
            return d.rename("data").to_dataset()

        # Raises on missing coords
        with pytest.raises(ValueError, match="This function adds the lon dimension*"):
            data = func(xr.Dataset(dict(tas=tas)), group="time.dayofyear", window=5)

        data = func(
            xr.Dataset(dict(tas=tas)),
            group="time.dayofyear",
            window=5,
            lon=[1, 2, 3, 4],
        ).load()
        assert set(data.data.dims) == {"time", "lon"}

    def test_grouper_prop(self, tas_series):
        tas = tas_series(np.arange(366), start="2000-01-01")
        tas = tas.expand_dims(lat=[1, 2, 3, 4]).chunk()

        @map_groups(data=[Grouper.PROP])
        def func(ds, *, dim):
            assert isinstance(dim, list)
            d = ds.tas.mean(dim)
            return d.rename("data").to_dataset()

        data = func(
            xr.Dataset(dict(tas=tas)),
            group="time.dayofyear",
            window=5,
            add_dims=["lat"],
        ).load()
        assert set(data.data.dims) == {"dayofyear"}

    def test_grouper_prop_main_only(self, tas_series):
        tas = tas_series(np.arange(366), start="2000-01-01")
        tas = tas.expand_dims(lat=[1, 2, 3, 4]).chunk()

        @map_groups(data=[Grouper.PROP], main_only=True)
        def func(ds, *, dim):
            assert isinstance(dim, str)
            data = ds.tas.mean(dim)
            return data.rename("data").to_dataset()

        # with a scalar aux coord
        data = func(
            xr.Dataset(dict(tas=tas.isel(lat=0, drop=True)), coords=dict(leftover=1)),
            group="time.dayofyear",
        ).load()
        assert set(data.data.dims) == {"dayofyear"}
        assert "leftover" in data

    def test_raises_error(self, tas_series):
        tas = tas_series(np.arange(366), start="2000-01-01")
        tas = tas.expand_dims(lat=[1, 2, 3, 4]).chunk(lat=1)

        # Test dim parsing
        @map_blocks(reduces=["lat"], data=[])
        def func(ds, *, group, lon=None):
            return ds.tas.rename("data").to_dataset()

        with pytest.raises(ValueError, match="cannot be chunked"):
            func(xr.Dataset(dict(tas=tas)), group="time")

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_dataarray_cfencode(self, use_dask, open_dataset):
        ds = open_dataset("sdba/CanESM2_1950-2100.nc")
        if use_dask:
            ds = ds.chunk()

        @map_blocks(reduces=["location"], data=[])
        def func(ds, *, group):
            d = ds.mean("location")
            return d.rename("data").to_dataset()

        with set_options(sdba_encode_cf=True):
            func(ds.convert_calendar("noleap").tasmax, group=Grouper("time"))



"""Precipitation indicator definitions."""

from __future__ import annotations

from xclim import indices
from xclim.core import cfchecks
from xclim.core.indicator import (
    Daily,
    Hourly,
    Indicator,
    ResamplingIndicator,
    ResamplingIndicatorWithIndexing,
)
from xclim.core.utils import InputKind

__all__ = [
    "cffwis_indices",
    "cold_and_dry_days",
    "cold_and_wet_days",
    "daily_pr_intensity",
    "days_over_precip_doy_thresh",
    "days_over_precip_thresh",
    "days_with_snow",
    "drought_code",
    "dry_days",
    "dry_spell_frequency",
    "dry_spell_max_length",
    "dry_spell_total_length",
    "dryness_index",
    "duff_moisture_code",
    "first_snowfall",
    "fraction_over_precip_doy_thresh",
    "fraction_over_precip_thresh",
    "griffiths_drought_factor",
    "high_precip_low_temp",
    "keetch_byram_drought_index",
    "last_snowfall",
    "liquid_precip_accumulation",
    "liquid_precip_average",
    "liquid_precip_ratio",
    "max_1day_precipitation_amount",
    "max_n_day_precipitation_amount",
    "max_pr_intensity",
    "maximum_consecutive_dry_days",
    "maximum_consecutive_wet_days",
    "mcarthur_forest_fire_danger_index",
    "precip_accumulation",
    "precip_average",
    "rain_on_frozen_ground_days",
    "rain_season",
    "rprctot",
    "snowfall_frequency",
    "snowfall_intensity",
    "solid_precip_accumulation",
    "solid_precip_average",
    "standardized_precipitation_evapotranspiration_index",
    "standardized_precipitation_index",
    "warm_and_dry_days",
    "warm_and_wet_days",
    "water_cycle_intensity",
    "wet_precip_accumulation",
    "wet_spell_frequency",
    "wet_spell_max_length",
    "wet_spell_total_length",
    "wetdays",
    "wetdays_prop",
]


class FireWeather(Indicator):
    """Non resampling - precipitation related indicators."""

    src_freq = "D"
    context = "hydro"
    keywords = "fire"


class Precip(Daily):
    """Indicator involving daily pr series."""

    context = "hydro"
    keywords = "precipitation"


class PrecipWithIndexing(ResamplingIndicatorWithIndexing):
    """Indicator involving daily pr series and allowing indexing."""

    src_freq = "D"
    context = "hydro"
    keywords = "precipitation"


class PrTasxWithIndexing(ResamplingIndicatorWithIndexing):
    """Indicator involving pr and one of tas, tasmin or tasmax, allowing indexing."""

    src_freq = "D"
    context = "hydro"
    keywords = "precipitation"

    def cfcheck(self, pr, tas):
        cfchecks.cfcheck_from_name("pr", pr)
        cfchecks.check_valid(tas, "standard_name", "air_temperature")


class StandardizedIndexes(ResamplingIndicator):
    """Resampling but flexible inputs indicators."""

    src_freq = ["D", "MS"]
    context = "hydro"


class HrPrecip(Hourly):
    """Indicator involving hourly pr series."""

    context = "hydro"
    keywords = "precipitation"


rain_on_frozen_ground_days = PrTasxWithIndexing(
    title="Number of rain on frozen ground days",
    identifier="rain_frzgr",
    units="days",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_above_threshold",
    long_name="Number of rain on frozen ground days (mean daily temperature > 0℃ and precipitation > {thresh})",
    description="{freq} number of days with rain above {thresh} after a series of seven days with average daily "
    "temperature below 0℃. Precipitation is assumed to be rain when the daily average temperature is above 0℃.",
    abstract="The number of days with rain above a given threshold after a series of seven days with average daily "
    "temperature below 0°C. Precipitation is assumed to be rain when the daily average temperature is above 0°C.",
    cell_methods="",
    compute=indices.rain_on_frozen_ground_days,
)

max_1day_precipitation_amount = PrecipWithIndexing(
    title="Maximum 1-day total precipitation",
    identifier="rx1day",
    units="mm/day",
    standard_name="lwe_thickness_of_precipitation_amount",
    long_name="Maximum 1-day total precipitation",
    description="{freq} maximum 1-day total precipitation",
    abstract="Maximum total daily precipitation for a given period.",
    cell_methods="time: maximum over days",
    compute=indices.max_1day_precipitation_amount,
)

max_n_day_precipitation_amount = Precip(
    title="maximum n-day total precipitation",
    identifier="max_n_day_precipitation_amount",
    var_name="rx{window}day",
    units="mm",
    standard_name="lwe_thickness_of_precipitation_amount",
    long_name="maximum {window}-day total precipitation amount",
    description="{freq} maximum {window}-day total precipitation amount.",
    abstract="Maximum of the moving sum of daily precipitation for a given period.",
    cell_methods="time: maximum over days",
    compute=indices.max_n_day_precipitation_amount,
)

wetdays = PrecipWithIndexing(
    title="Number of wet days",
    identifier="wetdays",
    units="days",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_at_or_above_threshold",
    long_name="Number of days with daily precipitation at or above {thresh}",
    description="{freq} number of days with daily precipitation at or above {thresh}.",
    abstract="The number of days with daily precipitation at or above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.wetdays,
)

wetdays_prop = PrecipWithIndexing(
    title="Proportion of wet days",
    identifier="wetdays_prop",
    units="1",
    long_name="Proportion of days with precipitation at or above {thresh}",
    description="{freq} proportion of days with precipitation at or above {thresh}.",
    abstract="The proportion of days with daily precipitation at or above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.wetdays_prop,
)

dry_days = PrecipWithIndexing(
    title="Number of dry days",
    identifier="dry_days",
    units="days",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_below_threshold",
    long_name="Number of dry days",
    description="{freq} number of days with daily precipitation under {thresh}.",
    abstract="The number of days with daily precipitation under a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.dry_days,
)

dryness_index = Precip(
    title="Dryness index",
    identifier="dryness_index",
    units="mm",
    long_name="Growing season humidity",
    description="Estimation of growing season humidity (precipitation minus adjusted evapotranspiration) for the "
    "period of April to September (Northern Hemisphere) or October to March (Southern Hemisphere), with initial soil "
    "moisture content set to {wo} and an adjustment based on monthly precipitation and evapotranspiration limits.",
    abstract="The dryness index is a characterization of the water component in winegrowing regions which considers "
    "the precipitation and evapotranspiration factors without deduction for surface runoff or drainage. "
    "Metric originally published in Riou et al. (1994).",
    cell_methods="",
    src_freq=["D", "MS"],
    compute=indices.dryness_index,
)

maximum_consecutive_wet_days = Precip(
    title="Maximum consecutive wet days",
    identifier="cwd",
    units="days",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_at_or_above_threshold",
    long_name="Maximum consecutive days with daily precipitation at or above {thresh}",
    description="{freq} maximum number of consecutive days with daily precipitation at or above {thresh}.",
    abstract="The longest number of consecutive days where daily precipitation is at or above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.maximum_consecutive_wet_days,
)

maximum_consecutive_dry_days = Precip(
    title="Maximum consecutive dry days",
    identifier="cdd",
    units="days",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_below_threshold",
    long_name="Maximum consecutive days with daily precipitation below {thresh}",
    description="{freq} maximum number of consecutive days with daily precipitation below {thresh}.",
    abstract="The longest number of consecutive days where daily precipitation below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.maximum_consecutive_dry_days,
)

daily_pr_intensity = PrecipWithIndexing(
    title="Simple Daily Intensity Index",
    identifier="sdii",
    units="mm d-1",
    standard_name="lwe_thickness_of_precipitation_amount",
    long_name="Average precipitation during days with daily precipitation over {thresh} "
    "(Simple Daily Intensity Index: SDII)",
    description="{freq} Simple Daily Intensity Index (SDII) or "
    "{freq} average precipitation for days with daily precipitation over {thresh}.",
    abstract="Average precipitation for days with daily precipitation above a given threshold.",
    cell_methods="",
    compute=indices.daily_pr_intensity,
)

max_pr_intensity = HrPrecip(
    title="Maximum precipitation intensity over time window",
    identifier="max_pr_intensity",
    units="mm h-1",
    standard_name="precipitation",
    long_name="Maximum precipitation intensity over rolling {window}h time window",
    description="{freq} maximum precipitation intensity over rolling {window}h time window.",
    abstract="Maximum precipitation intensity over a given rolling time window.",
    cell_methods="time: max",
    compute=indices.max_pr_intensity,
    duration="{window}",
    keywords="IDF curves",
)

precip_accumulation = PrecipWithIndexing(
    title="Total accumulated precipitation (solid and liquid)",
    identifier="prcptot",
    units="mm",
    standard_name="lwe_thickness_of_precipitation_amount",
    long_name="Total accumulated precipitation",
    description="{freq} total precipitation.",
    abstract="Total accumulated precipitation. If the average daily temperature is given, the phase parameter can be "
    "used to restrict the calculation to precipitation of only one phase (liquid or solid). Precipitation is "
    "considered solid if the average daily temperature is below 0°C (and vice versa).",
    cell_methods="time: sum over days",
    compute=indices.precip_accumulation,
    parameters={"tas": None, "phase": None},
)

precip_average = PrecipWithIndexing(
    title="Averaged precipitation (solid and liquid)",
    identifier="prcpavg",
    units="mm",
    standard_name="lwe_average_of_precipitation_amount",
    long_name="Averaged precipitation",
    description="{freq} mean precipitation.",
    abstract="Averaged precipitation. If the average daily temperature is given, the phase parameter can be "
    "used to restrict the calculation to precipitation of only one phase (liquid or solid). Precipitation is "
    "considered solid if the average daily temperature is below 0°C threshold (and vice versa).",
    cell_methods="time: mean over days",
    compute=indices.precip_average,
    parameters={"tas": None, "phase": None},
)

wet_precip_accumulation = PrecipWithIndexing(
    title="Total accumulated precipitation (solid and liquid) during wet days",
    identifier="wet_prcptot",
    units="mm",
    standard_name="lwe_thickness_of_precipitation_amount",
    long_name="Total accumulated precipitation over days where precipitation exceeds {thresh}",
    description="{freq} total precipitation over wet days, defined as days where precipitation exceeds {thresh}.",
    abstract="Total accumulated precipitation on days with precipitation. "
    "A day is considered to have precipitation if the precipitation is greater than or equal to a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.prcptot,
    parameters={"thresh": {"default": "1 mm/day"}},
)

liquid_precip_accumulation = PrTasxWithIndexing(
    title="Total accumulated liquid precipitation.",
    identifier="liquidprcptot",
    units="mm",
    standard_name="lwe_thickness_of_liquid_precipitation_amount",
    long_name="Total accumulated precipitation when temperature is above {thresh}",
    description="{freq} total {phase} precipitation, estimated as precipitation when temperature is above {thresh}.",
    abstract="Total accumulated liquid precipitation. "
    "Precipitation is considered liquid when the average daily temperature is above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.precip_accumulation,
    parameters={"tas": {"kind": InputKind.VARIABLE}, "phase": "liquid"},
)

liquid_precip_average = PrTasxWithIndexing(
    title="Averaged liquid precipitation.",
    identifier="liquidprcpavg",
    units="mm",
    standard_name="lwe_average_of_liquid_precipitation_amount",
    long_name="Averaged precipitation when temperature is above {thresh}",
    description="{freq} mean {phase} precipitation, estimated as precipitation when temperature is above {thresh}.",
    abstract="Averaged liquid precipitation. "
    "Precipitation is considered liquid when the average daily temperature is above a given threshold.",
    cell_methods="time: mean over days",
    compute=indices.precip_average,
    parameters={"tas": {"kind": InputKind.VARIABLE}, "phase": "liquid"},
)

solid_precip_accumulation = PrTasxWithIndexing(
    title="Total accumulated solid precipitation.",
    identifier="solidprcptot",
    units="mm",
    standard_name="lwe_thickness_of_snowfall_amount",
    long_name="Total accumulated solid precipitation",
    description="{freq} total solid precipitation, estimated as precipitation when temperature at or below {thresh}.",
    abstract="Total accumulated solid precipitation. "
    "Precipitation is considered solid when the average daily temperature is at or below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.precip_accumulation,
    parameters={"tas": {"kind": InputKind.VARIABLE}, "phase": "solid"},
)

solid_precip_average = PrTasxWithIndexing(
    title="Averaged solid precipitation.",
    identifier="solidprcpavg",
    units="mm",
    standard_name="lwe_average_of_snowfall_amount",
    long_name="Averaged solid precipitation",
    description="{freq} mean solid precipitation, estimated as precipitation when temperature at or below {thresh}.",
    abstract="Averaged solid precipitation. "
    "Precipitation is considered solid when the average daily temperature is at or below a given threshold.",
    cell_methods="time: mean over days",
    compute=indices.precip_average,
    parameters={"tas": {"kind": InputKind.VARIABLE}, "phase": "solid"},
)

standardized_precipitation_index = StandardizedIndexes(
    title="Standardized Precipitation Index (SPI)",
    identifier="spi",
    units="",
    standard_name="spi",
    long_name="Standardized Precipitation Index (SPI)",
    description="Precipitations over a moving {window}-X window, normalized such that SPI averages to 0 for "
    "calibration data. The window unit `X` is the minimal time period defined by resampling frequency {freq}.",
    abstract="Precipitation over a moving window, normalized such that SPI averages to 0 for the calibration data. "
    "The window unit `X` is the minimal time period defined by the resampling frequency.",
    cell_methods="",
    keywords="precipitation",
    compute=indices.standardized_precipitation_index,
)

standardized_precipitation_evapotranspiration_index = StandardizedIndexes(
    title="Standardized Precipitation Evapotranspiration Index (SPEI)",
    identifier="spei",
    units="",
    standard_name="spei",
    long_name="Standardized precipitation evapotranspiration index (SPEI)",
    description="Water budget (precipitation minus evapotranspiration) over a moving {window}-X window, normalized "
    "such that SPEI averages to 0 for calibration data. The window unit `X` is the minimal time period defined by the "
    "resampling frequency {freq}.",
    abstract="Water budget (precipitation - evapotranspiration) over a moving window, normalized such that the "
    "SPEI averages to 0 for the calibration data. The window unit `X` is the minimal time period defined by the "
    "resampling frequency.",
    cell_methods="",
    keywords="precipitation",
    compute=indices.standardized_precipitation_evapotranspiration_index,
)

drought_code = FireWeather(
    title="Daily drought code",
    identifier="dc",
    units="",
    long_name="Drought Code",
    description="Numerical code estimating the average moisture content of organic layers.",
    abstract="The Drought Index is part of the Canadian Forest-Weather Index system. "
    "It is a numerical code that estimates the average moisture content of organic layers.",
    compute=indices.drought_code,
    missing="skip",
)

duff_moisture_code = FireWeather(
    identifier="dmc",
    units="",
    long_name="Duff Moisture Code",
    description="Numeric rating of the average moisture content of loosely compacted organic layers of moderate depth.",
    compute=indices.duff_moisture_code,
    missing="skip",
)

cffwis_indices = FireWeather(
    identifier="cffwis",
    realm="atmos",
    var_name=["dc", "dmc", "ffmc", "isi", "bui", "fwi"],
    standard_name=[
        "drought_code",
        "duff_moisture_code",
        "fine_fuel_moisture_code",
        "initial_spread_index",
        "buildup_index",
        "fire_weather_index",
    ],
    long_name=[
        "Drought Code",
        "Duff Moisture Code",
        "Fine Fuel Moisture Code",
        "Initial Spread Index",
        "Buildup Index",
        "Fire Weather Index",
    ],
    description=[
        "Numeric rating of the average moisture content of deep, compact organic layers.",
        "Numeric rating of the average moisture content of loosely compacted organic layers of moderate depth.",
        "Numeric rating of the average moisture content of litter and other cured fine fuels.",
        "Numeric rating of the expected rate of fire spread.",
        "Numeric rating of the total amount of fuel available for combustion.",
        "Numeric rating of fire intensity.",
    ],
    units="",
    compute=indices.cffwis_indices,
    missing="skip",
)


keetch_byram_drought_index = FireWeather(
    identifier="kbdi",
    units="mm/day",
    standard_name="keetch_byram_drought_index",
    long_name="Keetch-Byran Drought Index",
    description="Amount of water necessary to bring the soil moisture content back to field capacity",
    compute=indices.keetch_byram_drought_index,
    missing="skip",
)


griffiths_drought_factor = FireWeather(
    identifier="df",
    units="",
    standard_name="griffiths_drought_factor",
    long_name="Griffiths Drought Factor",
    description="Numeric indicator of the forest fire fuel availability in the deep litter bed",
    compute=indices.griffiths_drought_factor,
    missing="skip",
)


mcarthur_forest_fire_danger_index = FireWeather(
    identifier="ffdi",
    units="",
    standard_name="mcarthur_forest_fire_danger_index",
    long_name="McArthur Forest Fire Danger Index",
    description="Numeric rating of the potential danger of a forest fire",
    compute=indices.mcarthur_forest_fire_danger_index,
    missing="skip",
)


last_snowfall = PrecipWithIndexing(
    title="Last day where snowfall exceeded a given threshold",
    identifier="last_snowfall",
    standard_name="day_of_year",
    long_name="Date of last day where snowfall exceeded {thresh}",
    description="{freq} last day where snowfall exceeded {thresh}.",
    abstract="The last day where snowfall exceeded a given threshold during a time period (the threshold can be "
    "given as a snowfall flux or a liquid water equivalent snowfall rate).",
    units="",
    compute=indices.last_snowfall,
)

first_snowfall = PrecipWithIndexing(
    title="First day where snowfall exceeded a given threshold",
    identifier="first_snowfall",
    standard_name="day_of_year",
    long_name="Date of first day where snowfall exceeded {thresh}",
    description="{freq} first day where snowfall exceeded {thresh}.",
    abstract="The first day where snowfall exceeded a given threshold during a time period (the threshold can be "
    "given as a snowfall flux or a liquid water equivalent snowfall rate).",
    units="",
    compute=indices.first_snowfall,
)

days_with_snow = PrecipWithIndexing(
    title="Days with snowfall",
    identifier="days_with_snow",
    long_name="Number of days with snowfall between {low} and {high} thresholds",
    description="{freq} number of days with snowfall larger than {low} and smaller or equal to {high}.",
    abstract="Number of days with snow between a lower and upper limit.",
    units="days",
    compute=indices.days_with_snow,
)

snowfall_frequency = PrecipWithIndexing(
    title="Snowfall frequency",
    identifier="snowfall_frequency",
    long_name="Percentage of days with snowfall above {thresh} threshold",
    description="{freq} percentage of days with snowfall larger than {thresh}.",
    abstract="Percentage of days with snowfall above a given threshold (either a "
    "snowfall flux or a liquid water equivalent snowfall rate).",
    units="%",
    compute=indices.snowfall_frequency,
)

snowfall_intensity = PrecipWithIndexing(
    title="Snowfall intensity",
    identifier="snowfall_intensity",
    long_name="Mean daily snowfall above {thresh} threshold",
    description="{freq} mean daily snowfall larger than {thresh}.",
    abstract="Mean daily liquid water equivalent snowfall rate above threshold (either a "
    "snowfall flux or a liquid water equivalent snowfall rate)",
    units="mm/day",
    compute=indices.snowfall_intensity,
)

# FIXME: Are days_over_precip_thresh and days_over_precip_doy_thresh the same thing?
days_over_precip_thresh = PrecipWithIndexing(
    title="Number of days with precipitation above a given percentile",
    identifier="days_over_precip_thresh",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_above_threshold",
    long_name="Number of days with precipitation flux above the {pr_per_thresh}th percentile of {pr_per_period}",
    description="{freq} number of days with precipitation above the {pr_per_thresh}th percentile of {pr_per_period} "
    "period. Only days with at least {thresh} are counted.",
    abstract="Number of days in a period where precipitation is above a given percentile, "
    "calculated over a given period and a fixed threshold.",
    units="days",
    cell_methods="time: sum over days",
    compute=indices.days_over_precip_thresh,
)

# FIXME: Are days_over_precip_thresh and days_over_precip_doy_thresh the same thing?
days_over_precip_doy_thresh = PrecipWithIndexing(
    title="Number of days with precipitation above a given daily percentile",
    identifier="days_over_precip_doy_thresh",
    standard_name="number_of_days_with_lwe_thickness_of_precipitation_amount_above_daily_threshold",
    long_name="Number of days with daily precipitation flux above the {pr_per_thresh}th percentile of {pr_per_period}",
    description="{freq} number of days with precipitation above the {pr_per_thresh}th daily percentile. Only days with "
    "at least {thresh} are counted. A {pr_per_window} day(s) window, centered on each calendar day in the "
    "{pr_per_period} period, is used to compute the {pr_per_thresh}th percentile(s).",
    abstract="Number of days in a period where precipitation is above a given daily percentile and a fixed threshold.",
    units="days",
    cell_methods="time: sum over days",
    compute=indices.days_over_precip_thresh,
)

high_precip_low_temp = PrTasxWithIndexing(
    title="Days with precipitation and cold temperature",
    identifier="high_precip_low_temp",
    long_name="Days with precipitation at or above {pr_thresh} and temperature below {tas_thresh}",
    description="{freq} number of days with precipitation at or above {pr_thresh} and temperature below {tas_thresh}.",
    abstract="Number of days with precipitation above a given threshold and temperature below a given threshold.",
    units="days",
    cell_methods="time: sum over days",
    compute=indices.high_precip_low_temp,
)

# FIXME: Are fraction_over_precip_thresh and fraction_over_precip_doy_thresh the same thing?
# FIXME: Clarity needed in both French and English metadata fields
fraction_over_precip_doy_thresh = PrecipWithIndexing(
    title="Fraction of precipitation due to wet days with daily precipitation over a given daily percentile.",
    identifier="fraction_over_precip_doy_thresh",
    long_name="Fraction of precipitation due to days with daily precipitation above {pr_per_thresh}th daily percentile",
    description="{freq} fraction of total precipitation due to days with precipitation above {pr_per_thresh}th daily "
    "percentile. Only days with at least {thresh} are included in the total. A {pr_per_window} day(s) window, centered "
    "on each calendar day in the {pr_per_period} period, is used to compute the {pr_per_thresh}th percentile(s).",
    units="",
    cell_methods="",
    compute=indices.fraction_over_precip_thresh,
)

# FIXME: Are fraction_over_precip_thresh and fraction_over_precip_doy_thresh the same thing?
# FIXME: Clarity needed in both French and English metadata fields
fraction_over_precip_thresh = PrecipWithIndexing(
    identifier="fraction_over_precip_thresh",
    long_name="Fraction of precipitation due to days with precipitation above {pr_per_thresh}th daily percentile",
    description="{freq} fraction of total precipitation due to days with precipitation above {pr_per_thresh}th "
    "percentile of {pr_per_period} period. Only days with at least {thresh} are included in the total.",
    units="",
    cell_methods="",
    compute=indices.fraction_over_precip_thresh,
)

liquid_precip_ratio = PrTasxWithIndexing(
    title="Fraction of liquid to total precipitation",
    identifier="liquid_precip_ratio",
    long_name="Fraction of liquid to total precipitation (temperature above {thresh})",
    description="The {freq} ratio of rainfall to total precipitation. Rainfall is estimated as precipitation on days "
    "where temperature is above {thresh}.",
    abstract="The ratio of total liquid precipitation over the total precipitation. Liquid precipitation is "
    "approximated from total precipitation on days where temperature is above a given threshold.",
    units="",
    compute=indices.liquid_precip_ratio,
    parameters={"tas": {"kind": InputKind.VARIABLE}, "prsn": None},
)


dry_spell_frequency = Precip(
    title="Dry spell frequency",
    identifier="dry_spell_frequency",
    long_name="Number of dry periods of {window} day(s) or more, during which the {op} precipitation on a "
    "window of {window} day(s) is below {thresh}.",
    description="The {freq} number of dry periods of {window} day(s) or more, during which the {op} precipitation on a "
    "window of {window} day(s) is below {thresh}.",
    abstract="The frequency of dry periods of `N` days or more, during which the accumulated or maximum precipitation "
    "over a given time window of days is below a given threshold.",
    units="",
    cell_methods="",
    compute=indices.dry_spell_frequency,
)


dry_spell_total_length = Precip(
    title="Dry spell total length",
    identifier="dry_spell_total_length",
    long_name="Number of days in dry periods of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is under {thresh}.",
    description="The {freq} number of days in dry periods of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is under {thresh}.",
    abstract="The total length of dry periods of `N` days or more, during which the accumulated or maximum "
    "precipitation over a given time window of days is below a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.dry_spell_total_length,
)

dry_spell_max_length = Precip(
    title="Dry spell maximum length",
    identifier="dry_spell_max_length",
    long_name="Maximum consecutive number of days in a dry period of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is under {thresh}.",
    description="The maximum {freq} number of consecutive days in a dry period of {window} day(s) or more"
    ", during which the {op} precipitation within windows of {window} day(s) is under {thresh}.",
    abstract="The maximum length of a dry period of `N` days or more, during which the accumulated or maximum "
    "precipitation over a given time window of days is below a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.dry_spell_max_length,
)

wet_spell_frequency = Precip(
    title="Wet spell frequency",
    identifier="wet_spell_frequency",
    long_name="Number of wet periods of {window} day(s) or more, during which the {op} precipitation on a "
    "window of {window} day(s) is equal or over {thresh}.",
    description="The {freq} number of wet periods of {window} day(s) or more, during which the {op} precipitation on a "
    "window of {window} day(s) is equal or over {thresh}.",
    abstract="The frequency of wet periods of `N` days or more, during which the accumulated or maximum precipitation "
    "over a given time window of days is equal or above a given threshold.",
    units="",
    cell_methods="",
    compute=indices.wet_spell_frequency,
)


wet_spell_total_length = Precip(
    title="Wet spell total length",
    identifier="wet_spell_total_length",
    long_name="Number of days in wet periods of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is equal or over {thresh}.",
    description="The {freq} number of days in wet periods of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is equal or over {thresh}.",
    abstract="The total length of dry periods of `N` days or more, during which the accumulated or maximum "
    "precipitation over a given time window of days is equal or above a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.wet_spell_total_length,
)

wet_spell_max_length = Precip(
    title="Wet spell maximum length",
    identifier="wet_spell_max_length",
    long_name="Maximum consecutive number of days in a wet period of {window} day(s) or more, during which the {op} "
    "precipitation within windows of {window} day(s) is equal or over {thresh}.",
    description="The maximum {freq} number of consecutive days in a wet period of {window} day(s) or more"
    ", during which the {op} precipitation within windows of {window} day(s) is equal or over {thresh}.",
    abstract="The maximum length of a wet period of `N` days or more, during which the accumulated or maximum "
    "precipitation over a given time window of days is equal or above a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.wet_spell_max_length,
)

rprctot = PrecipWithIndexing(
    title="Proportion of accumulated precipitation arising from convective processes",
    identifier="rprctot",
    long_name="Proportion of accumulated precipitation arising from convective processes"
    "with precipitation of at least {thresh}",
    description="{freq} proportion of accumulated precipitation arising from convective processes "
    "with precipitation of at least {thresh}.",
    abstract="The proportion of total precipitation due to convective processes. "
    "Only days with surpassing a minimum precipitation flux are considered.",
    units="",
    cell_methods="time: sum",
    compute=indices.rprctot,
)


cold_and_dry_days = PrecipWithIndexing(
    title="Cold and dry days",
    identifier="cold_and_dry_days",
    units="days",
    long_name="Number of days where temperature is below {tas_per_thresh}th percentile and "
    "precipitation is below {pr_per_thresh}th percentile",
    description="{freq} number of days where temperature is below {tas_per_thresh}th percentile and "
    "precipitation is below {pr_per_thresh}th percentile.",
    abstract="Number of days with temperature below a given percentile and precipitation below a given percentile.",
    cell_methods="time: sum over days",
    compute=indices.cold_and_dry_days,
)

warm_and_dry_days = PrecipWithIndexing(
    title="Warm and dry days",
    identifier="warm_and_dry_days",
    units="days",
    long_name="Number of days where temperature is above {tas_per_thresh}th percentile and "
    "precipitation is below {pr_per_thresh}th percentile",
    description="{freq} number of days where temperature is above {tas_per_thresh}th percentile and "
    "precipitation is below {pr_per_thresh}th percentile.",
    abstract="Number of days with temperature above a given percentile and precipitation below a given percentile.",
    cell_methods="time: sum over days",
    compute=indices.warm_and_dry_days,
)

warm_and_wet_days = PrecipWithIndexing(
    title="Warm and wet days",
    identifier="warm_and_wet_days",
    units="days",
    long_name="Number of days where temperature above {tas_per_thresh}th percentile and "
    "precipitation above {pr_per_thresh}th percentile",
    description="{freq} number of days where temperature is above {tas_per_thresh}th percentile and "
    "precipitation is above {pr_per_thresh}th percentile.",
    abstract="Number of days with temperature above a given percentile and precipitation above a given percentile.",
    cell_methods="time: sum over days",
    compute=indices.warm_and_wet_days,
)

cold_and_wet_days = PrecipWithIndexing(
    title="Cold and wet days",
    identifier="cold_and_wet_days",
    units="days",
    long_name="Number of days where temperature is below {tas_per_thresh}th percentile and "
    "precipitation is above {pr_per_thresh}th percentile",
    description="{freq} number of days where temperature is below {tas_per_thresh}th percentile and "
    "precipitation is above {pr_per_thresh}th percentile.",
    abstract="Number of days with temperature below a given percentile and precipitation above a given percentile.",
    cell_methods="time: sum over days",
    compute=indices.cold_and_wet_days,
)

rain_season = Precip(
    title="Rain season",
    identifier="rain_season",
    realm="atmos",
    var_name=["rain_season_start", "rain_season_end", "rain_season_length"],
    long_name=[
        "Start of the rain season",
        "End of the rain season",
        "Length of the rain season",
    ],
    description=[
        "First step of a run where i) a sequence of {window_wet_start} days accumulated {thresh_wet_start} "
        "of precipitations ii) followed by a sequence of {window_not_dry_start} days with no dry sequence, "
        "i.e. a sequence of {window_dry_start} days with at least {thresh_dry_start} {method_dry_start}. "
        "The start of the season is on the last day of the first sequence i) and must be between {date_min_start} and "
        "{date_max_start}.",
        "Last day in a dry sequence after the start of the season, i.e. a sequence of {window_dry_end} days "
        "with at least {thresh_dry_end} {method_dry_end}. It must be between {date_min_end} and {date_max_end}. ",
        "Number of steps of the original series in the season, between 'start' and 'end'.",
    ],
    units=["", "", "days"],
    abstract="Start time, end time and length of the rain season, notably useful for West Africa (sivakumar, 1998). "
    "The rain season starts with a period of abundant rainfall, followed by a period without prolonged dry sequences, "
    "which must happen before a given date. The rain season stops during a dry period happening after a given date.",
    cell_methods="",
    compute=indices.rain_season,
)

water_cycle_intensity = PrecipWithIndexing(
    title="Water cycle intensity",
    identifier="water_cycle_intensity",
    realm="atmos",
    units="mm",
    long_name="Water cycle intensity",
    description="The {freq} water cycle intensity, defined as the sum of precipitation and actual evapotranspiration.",
    abstract="The sum of precipitation and actual evapotranspiration.",
    cell_methods="time: sum over days",
    compute=indices.water_cycle_intensity,
)
